// //////////////////////////////////////////////////////////////////////////
//
//     compact_map3d.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#ifndef _COMPACT_MAP3D_H_
#define _COMPACT_MAP3D_H_ 1

#include "crystal.h"

////////////////////////////
//
// map_3d template class
//
////////////////////////////
template <class T>
class compact_map_3d {
public:
	compact_map_3d() {
		//cout << "INITIALIZING() " << (unsigned int)this << "\n";
		data = NULL;
		ptrMap = NULL; invPtrMap = NULL;
		nsymm = 0;
		dims[0]=dims[1]=dims[2]=0;
		nPts = 0;
	}
	~compact_map_3d() {
		//cout << "DELETING " << (unsigned int)this << "\n";
		if (data) delete[] data; data = NULL;
	}

	// copy constructor
	compact_map_3d(const compact_map_3d<T>& X) {
		//cout << "COPYING [" << X.ptrMap << "," << X.invPtrMap << "," << X.nsymm << "," << X.symmGpId << ","
		//     << X.dims[0] << "," << X.dims[1] << "," << X.dims[2] << "\n";
		data = NULL; nPts = 0;
		if (X.nPts > 0) {
			alloc(X.ptrMap, X.invPtrMap, X.nsymm, X.symmGpId, X.dims[0], X.dims[1], X.dims[2]);
			memcpy(data, X.data, nPts*sizeof(T));
		} else {
			clear();
		}
	}
	compact_map_3d<T>& operator=(const compact_map_3d<T>& X) {
		//cout << "ASSIGNING [" << X.ptrMap << "," << X.invPtrMap << "," << X.nsymm << "," << X.symmGpId << ","
		//     << X.dims[0] << "," << X.dims[1] << "," << X.dims[2] << "\n";

		if (X.nPts > 0) {
			alloc(X.ptrMap, X.invPtrMap, X.nsymm, X.symmGpId, X.dims[0], X.dims[1], X.dims[2]);
												// only reallocates IF NECESSARY
			memcpy(data, X.data, nPts*sizeof(T));
		} else {
			clear();
		}
		return *this;
	}

	// (re)allocate storage space for this map
	void alloc(Crystal &crystData, vec3<int> X) {
		alloc(crystData, X[0],X[1],X[2]);
	}
	void alloc(Crystal &crystData, int x, int y, int z) {
		if (x != dims[0] || y != dims[1] || z != dims[2] || symmGpId != crystData.symmGpId ) {   // only reallocate if necessary
			if (data) {
				delete data; data = NULL;
			}

			nsymm     =  crystData.nSymm();
			ptrMap    = &crystData.getPtrMap();
			invPtrMap = &crystData.getPtrMapInv();
			symmGpId  =  crystData.symmGpId;

			nPts = invPtrMap->size();

			data = new T[nPts];
			dims[0]=x; dims[1]=y; dims[2]=z;

			// error check
			if (data == NULL)
				cerr << "ERROR: map_3d::alloc() returned NULL\n";
		}
	}
	void alloc(map_3d<int> *myPtrMap, vector< vec3<int> > *myInvPtrMap, int myNsymm, string mySymmGpId, int x, int y, int z) {
		if (x != dims[0] || y != dims[1] || z != dims[2] || symmGpId != mySymmGpId ) {   // only reallocate if necessary
			if (data) {
				delete data; data = NULL;
			}

			nsymm     = myNsymm;
			ptrMap    = myPtrMap;
			invPtrMap = myInvPtrMap;
			symmGpId  = mySymmGpId;

			nPts = invPtrMap->size();
			data = new T[nPts];
			dims[0]=x; dims[1]=y; dims[2]=z;

			// error check
			if (data == NULL)
				cerr << "ERROR: map_3d::alloc() returned NULL\n";
		}
	}
	void clear() {
		if (data) {
			delete data; data = NULL;
		}
		ptrMap = NULL; invPtrMap = NULL;
		dims[0]=dims[1]=dims[2]=0;
		nPts = 0;
		nsymm = 0;
	}

	//
	// expand (and recast) to a full map
	//
	template<class S>
	void inflate(map_3d<S> &full) {
		full.alloc(dims[0],dims[1],dims[2]);

		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; ++i) {
			full.set(i, (S)i_get(i));
		}
	}

	//
	// dumpDataToDisk
	//      make a raw dump of 'data' to disk, freeing up memory
	//
	void dumpDataToDisk(string filename) {
		writeDataToDisk(filename);
		dumpNoWrite();
	}
	//
	// writeDataToDisk
	//      make a raw dump of 'data' to disk, but keep in memory
	//
	void writeDataToDisk(string filename) {
		//return;
		if (nPts == 0) {
			cerr<< "ERROR: trying to dump empty map " << filename << endl;
			return;
		}
		if (data == NULL) {
			cerr<< "ERROR: trying to dump null map " << filename << endl;
			return;
		}
		ofstream out;

		out.open(filename.c_str(), ios::out);
		if (!out.is_open())  {
			cerr << "ERROR: Unable to open datfile '" << filename << "'\n";
			return;
		}

		for(int i = 0; i < nPts; i++){
			T buff = data[i];
			out.write((char*)&buff, sizeof(T));
			if(!out.good()){
				cerr << "Fail at "<< i << ", " << filename << endl;
				break;
			}
		}
		out.close();
	}
	//
	void dumpNoWrite() {
		//return;
		if (nPts == 0) {
			cerr<< "ERROR: deallocating empty data!" << endl;
			return;
		}
		if (data == NULL) {
			cerr<< "ERROR: deallocating null data!" << endl;
			return;
		}

		delete[] data; data = NULL;
		//cout << "DUMP NULL\n";
	}


	//
	// loadDataFromDisk
	//      read in a previous dump of 'data'
	//
	void loadDataFromDisk(string filename) {
		//return;
		if (nPts == 0) {
			cerr<< "ERROR: trying to load null map " << filename << endl;
			return;
		}
		if (data != NULL) {
			cerr<< "WARNING: loading " << filename << " would overwrite data!" << endl;
			delete[] data; data = NULL;
			//return;
		}

		ifstream in;
 		in.open(filename.c_str(), ios::in );

		if (!in.is_open())  {
			cerr << "ERROR: Unable to open datfile '" << filename << "'\n";
			return;
		}

		data = new T[nPts];
		in.read((char*)data, sizeof(T)*nPts);

//		for (int i=0; i<dims[0]; ++i) {
//			for (int j=0; j<dims[1]; ++j) {
//				for (int k=0; k<dims[2]; ++k) {
//					int idx = ptrMap->get(i,j,k);
//					if (data[idx] != datab[idx])
//						cout << "     !! " << filename << "!! " << idx << " (" << i << "," <<j<<","<<k<<")\n";
//				}
//			}
//		}

		//cout << "LOAD " << filename << " (" << data[nPts-1] << ")" << endl;
		in.close();
	}

	//
	void realloc() {
		//return;
		if (nPts == 0) {
			cerr<< "ERROR: trying to reallocate null map" << endl;
			return;
		}
		if (data != NULL) {
			cerr<< "WARNING: reallocating non-NULL data!" << endl;
			delete[] data; data = NULL;
			//return;
		}

		data = new T[nPts];
		// error check
		if (data == NULL)
			cerr << "ERROR: map_3d::alloc() returned NULL\n";

		//cout << "REALLOC NULL\n";
	}

	void testIfContainsNaN(string id) {
		for (int i=0; i<nPts; i++)
			if (!isfinite(data[i])) {
				cerr << id << "! NaN at " << i << " (" << data[i] << ")\n";
				exit(1);
			}
	}
	//
	// sum
	T sum() {
		T retval=0;
		for (int i=0; i<nPts; ++i)
			retval += data[i];
		return (retval);
	}

	T sum2() {
		T retval=0;
		for (int i=0; i<nPts; ++i)
			retval += data[i]*data[i];
		return (retval);
	}

	//
	// mean/variance
	T mean() {
		T sum=0;
		for (int i=0; i<nPts; ++i) {
			sum += data[i];
		}
		return ( sum / (double)nPts );
	}
	T variance() {
		T sum=0, sum2=0;
		for (int i=0; i<nPts; ++i) {
			sum += data[i];
			sum2 += data[i]*data[i];
		}
		double N2 = square((double)nPts);
		return ( (nPts*sum2-sum*sum) /  (N2) );
	}

	//
	// max
	void argmax(T& maxX, vec3<int> &argmaxX) {
		maxX = data[0];
		int idxMax = 0;
		for (int i=1; i<nPts; i++) {
			if (maxX < data[i]) {
				maxX = data[i];
				idxMax = i;
			}
		}
		argmaxX = (*invPtrMap)[idxMax];
	}
	T getMax() {
		if(nPts==0) return -1;
		T maxX = data[0];
		for (int i=1; i<nPts; i++) {
			if (maxX < data[i]) {
				maxX = data[i];
			}
		}
		return maxX;
	}

	//
	// information content
	T entropy() {
		T retval=(T)0;
		normalize();

		for (int i=0; i<nPts; i++)
			retval -= data[i] * (log(data[i])/log(2.0));

		return retval;
	}

	//
	// random location based on probabilities specified in map
	//   assume normalized, and 'invlog_transform'ed
	double randomLocation(vec3<int> &X) {
		vec3<double> fC;
		double randomIdx = ((double)rand())/RAND_MAX;

		for (int i=0; i<nPts; i++) {
			randomIdx -= data[i];
			if (randomIdx<=0) {
				X = (*invPtrMap)[i];
				return data[i];
			}
		}
		return -1;
	}

	//
	// map transformation functions
	void sqrtMap() {
		for (int i=0; i<nPts; i++)
			data[i] = sqrt(data[i]);
	}
	void invMap() {
		for (int i=0; i<nPts; i++)
			data[i] = 1/(data[i]);
	}


	// normalize map
	void normalize(double Mest=0.0) {
		double dataSum=0;
		for (int i=0; i<nPts; i++) {
			dataSum += data[i];
		}

		double Mperpoint = Mest / nPts;
		for (int i=0; i<nPts; i++)
			data[i] = (1-Mest) * data[i]/dataSum + Mperpoint;
	}

	// normalize log map, ensuring no value goes below log(Mest)
	void logNormalize(double Mest=0.0) {
		double maxValue=data[0];
		for (int i=1; i<nPts; i++)
			maxValue = max((T)data[i],(T)maxValue);

		// shift data by maxValue (==> multiplying by a constant to prevent underflow)
		// then take exp and sum
		double sum = 0;
		for (int i=0; i<nPts; i++) {
			data[i] = exp(data[i] - maxValue);
			sum += data[i];
		}

		double Mperpoint = Mest / nPts;
		for (int i=0; i<nPts; i++)
			data[i] = log( (1-Mest)*data[i] / sum + Mperpoint );
	}

	// transform to and from log space
	void log_transform() {
		for (int i=0; i<nPts; i++) {
			data[i] = log(data[i]);
		}
	}
	void invlog_transform() {
		for (int i=0; i<nPts; i++)
			data[i] = exp(data[i]);
	}

	// multiply map by a constant factor
	void scale(double factor) {
		for (int i=0; i<nPts; i++)
			data[i] *= factor;
	}

	// multiply map by a constant factor, then shift by shift
	void scaleShift(double factor, T shift) {
		for (int i=0; i<nPts; i++)
			data[i] = data[i]*factor + shift;
	}

	//
	// accessor functions
	//
	//  INDIRECT
	inline int idx_get(int x, int y, int z) {
		return ptrMap->get(x,y,z);
	}
	inline int idx_get(int xyz) {
		return ptrMap->get(xyz);
	}

	inline T& i_get(int x, int y, int z) const {
		//if ( x<0 || x>=dims[0] || y<0 || y>=dims[1] || z<0 || z>=dims[2] ) {
		//	cerr << "ERROR:  compact_map_3d::i_get() out of range access " << x <<"," << y << "," << z << endl;
		//	return 0;
		//}
		int idx = ptrMap->get(x,y,z);
		//if (idx<0 || idx>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::i_get() out of range pointer! " << idx << endl;
		//	return 0;
		//}
		return (data[ idx ]);
	}
	inline T& i_get(int xyz) const {
		//if ( xyz<0 || xyz>=dims[0]*dims[1]*dims[2] ) {
		//	cerr << "ERROR:  compact_map_3d::i_get() out of range access " << xyz << endl;
		//	return data[0];
		//}
		int idx = ptrMap->get(xyz);
		//if (idx<0 || idx>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::i_get() out of range pointer! " << idx << endl;
		//	return data[0];
		//}
		return (data[ idx ]);
	}

	inline void i_set(int x, int y, int z, T val) {
		//if ( x<0 || x>=dims[0] || y<0 || y>=dims[1] || z<0 || z>=dims[2] ) {
		//	cerr << "ERROR:  compact_map_3d::i_set() out of range access " << x <<"," << y << "," << z << endl;
		//	return;
		//}
		int idx = ptrMap->get(x,y,z);
		//if (idx<0 || idx>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::i_set() out of range pointer! " << idx << endl;
		//	return;
		//}
		data[idx] = val;
	}
	inline void i_set(int xyz, T val) {
		//if ( xyz<0 || xyz>=dims[0]*dims[1]*dims[2] ) {
		//	cerr << "ERROR:  compact_map_3d::i_set() out of range access " << xyz << endl;
		//	return;
		//}
		int idx = ptrMap->get(xyz);
		//if (idx<0 || idx>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::i_set() out of range pointer! " << idx << endl;
		//	return;
		//}
		data[idx] = val;
	}

	inline void i_inc(int x, int y, int z, T val) {
		//if ( x<0 || x>=dims[0] || y<0 || y>=dims[1] || z<0 || z>=dims[2] ) {
		//	cerr << "ERROR:  compact_map_3d::i_inc() out of range access " << x <<"," << y << "," << z << endl;
		//	return;
		//}
		int idx = ptrMap->get(x,y,z);
		//if (idx<0 || idx>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::i_inc() out of range pointer! " << idx << endl;
		//	return;
		//}
		data[idx] += val;
	}
	inline void i_inc(int xyz, T val) {
		//if ( xyz<0 || xyz>=dims[0]*dims[1]*dims[2] ) {
		//	cerr << "ERROR:  compact_map_3d::i_inc() out of range access " << xyz << endl;
		//	return;
		//}
		int idx = ptrMap->get(xyz);
		//if (idx<0 || idx>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::i_inc() out of range pointer! " << idx << endl;
		//	return;
		//}
		data[idx] += val;
	}

	//  DIRECT
	inline T&   d_get(int i)  const {
		//if (i<0 || i>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::d_get() out of range access " << i << endl;
		//	return data[0];
		//}
		return (data[i]);
	}

	inline void d_set(int i, T val) {
		//if (i<0 || i>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::d_set() out of range access " << i << endl;
		//	return;
		//}
		data[i] = val;
	}
	inline void d_inc(int i, T val) {
		//if (i<0 || i>=nPts) {
		//	cerr << "ERROR:  compact_map_3d::d_inc() out of range access " << i << endl;
		//	return;
		//}
		data[i] += val;
	}

	void setAll(T val) {
		for (int i=0; i < nPts; ++i) data[i] = val;
	}

	vec3<int> getDims() { return dims; }
	int getDims(int i)  { return dims[i]; }
	int getNPts()       { return nPts; }

private:
	vec3<int> dims;

	T* data;
	int nPts;

	map_3d<int> *ptrMap;
	vector< vec3<int> > *invPtrMap;
	int    nsymm;
	string symmGpId;
};


template <class S>
S map_dot(compact_map_3d<S>& A, compact_map_3d<S>& B) {
	S dotProd = 0.0;
	for (int i=0; i < A.getNPts(); i++)
		dotProd += A.d_get(i) * B.d_get(i);
	return dotProd;
}

template <class S>
S map_delta(compact_map_3d<S>& A, compact_map_3d<S>& B) {
	S dotProd = 0.0;
	for (int i=0; i < A.getNPts(); i++)
		dotProd += fabs( exp(A.d_get(i)) - exp(B.d_get(i)) );
	return dotProd;
}

// element-by-element multiplication of A and B gives AB
template <class T, class U>
void map_times(compact_map_3d<T>& AB, compact_map_3d<T>& A, compact_map_3d<U>& B) {
	int nPtsA = A.getNPts(), nPtsB = B.getNPts();
	if (nPtsA != nPtsB) {
		cerr << "compact_map_3d::map_times(): map " << nPtsA << " and map "
		                                            << nPtsB << " dimensions do not match!\n";
		return;
	}
	AB = A;
	for (int i=0; i < nPtsA; i++)
		AB.d_set(i , A.d_get(i) * B.d_get(i));
}

template <class T, class U>
void map_plus(compact_map_3d<T>& AB, compact_map_3d<T>& A, compact_map_3d<U>& B) {
	int nPtsA = A.getNPts(), nPtsB = B.getNPts();
	if (nPtsA != nPtsB) {
		cerr << "compact_map_3d::map_plus(): map " << nPtsA << " and map "
		                                            << nPtsB << " dimensions do not match!\n";
		return;
	}
	AB = A;
	for (int i=0; i < nPtsA; i++)
		AB.d_set(i , A.d_get(i) + B.d_get(i));
}

template <class T, class U>
void map_max(compact_map_3d<T>& AB, compact_map_3d<T>& A, compact_map_3d<U>& B) {
	int nPtsA = A.getNPts(), nPtsB = B.getNPts();
	if (nPtsA != nPtsB) {
		cerr << "compact_map_3d::map_plus(): map " << nPtsA << " and map "
		                                            << nPtsB << " dimensions do not match!\n";
		return;
	}
	AB = A;
	for (int i=0; i < nPtsA; i++)
		AB.d_set(i , (A.d_get(i) > B.d_get(i)) ? A.d_get(i) : B.d_get(i));
}

template <class T, class U>
void map_minus(compact_map_3d<T>& AB, compact_map_3d<T>& A, compact_map_3d<U>& B) {
	int nPtsA = A.getNPts(), nPtsB = B.getNPts();
	if (nPtsA != nPtsB) {
		cerr << "compact_map_3d::map_minus(): map " << nPtsA << " and map "
		                                            << nPtsB << " dimensions do not match!\n";
		return;
	}
	AB = A;
	for (int i=0; i < nPtsA; i++)
		AB.d_set(i , A.d_get(i) - B.d_get(i));
}

// element-by-element linear combination x*A + y*B gives AB
template <class T, class U>
void map_linear_combine(compact_map_3d<T>& AB, double x, compact_map_3d<T>& A, double y,  compact_map_3d<U>& B) {
	int nPtsA = A.getNPts(), nPtsB = B.getNPts();
	if (nPtsA != nPtsB) {
		cerr << "compact_map_3d::map_linear_combine(): map " << nPtsA << " and map "
		                                            << nPtsB << " dimensions do not match!\n";
		return;
	}
	AB = A;
	for (int i=0; i < nPtsA; i++)
		AB.d_set(i , x*A.d_get(i) + y*B.d_get(i));
}

//
//  write a map to disk
//
template <class S>
void writeMapToDisk(compact_map_3d<S> &M, string filestem) {
	map_3d<S> Mfull;
	M.inflate(Mfull);
	writeMapToDisk(Mfull, filestem);
}

//
// resize a compact map
//    by expandnig, fft-resizing, and then compacting
template <class S>
void resize_fft(compact_map_3d<S>& oldMap, compact_map_3d<S>& newMap) {
	vec3<int> olddims = oldMap.getDims();
	vec3<int> newdims = newMap.getDims();
	map_3d<double> oldFullMap(olddims), newFullMap(newdims);

	// expand
	int N = olddims[0]*olddims[1]*olddims[2];
	for (int i=0; i<N; ++i) {
		oldFullMap.set(i, (double)oldMap.i_get(i));
	}

	// fft
	resize_fft(oldFullMap,newFullMap);

	// contract
	N = newdims[0]*newdims[1]*newdims[2];
	for (int i=0; i<N; ++i) {
		newMap.i_set(i, (double)newFullMap.get(i));
	}
}

#endif

