// //////////////////////////////////////////////////////////////////////////
//
//     crystal.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#ifndef _CRYST_H_
#define _CRYST_H_ 1

#include "symmops.h"
#include "map3d.h"


void version(){
	string version = "$HeadURL: file:///u/soni/acmi/svn/tags/releases/1.5.1/common/crystal.h $";
	int begin = version.find("releases") + 9;
	int end = version.find("common")-1;

	cerr << "ACMI release version " << version.substr(begin,end-begin) << endl;
}

inline double d2r(double d) { return (d*M_PI/180.0); }

bool factorsLTE19(int X) {
	while (X != 1 && X%2 == 0) X /= 2;
	while (X != 1 && X%3 == 0) X /= 3;
	while (X != 1 && X%5 == 0) X /= 5;
	while (X != 1 && X%7 == 0) X /= 7;
	while (X != 1 && X%11 == 0)	X /= 11;
	while (X != 1 && X%13 == 0)	X /= 13;
	while (X != 1 && X%17 == 0)	X /= 17;
	while (X != 1 && X%19 == 0)	X /= 19;

	return (X == 1);
}

// modded fndsmp.f in CCP4
int findSampling(double MINSMP, int NMUL) {
	if (MINSMP <= 0) return NMUL;

	// multiple of nmul nearest minsmp
	int N = (int) floor( MINSMP/NMUL + 0.5 ) * NMUL;

	// increment until no factors >= 19
	while (!factorsLTE19(N))
		N += NMUL;

	return N;
}

class spacegroup {
private:
	vector< map_3d<int> > symmMaps;
	vector< R3_transform >  symmOps;
	vector< mat33<double> > symmInvRot;

	map_3d<short> asymmUnit;
	map_3d<short> asymmCopy;
	map_3d<int> ptrMap;     // each entry points to asymmUnit index
	vector< vec3<int> > ptrMapInv;  // map asymm index -> pt

	int Nsymm, mySpaceGpID;
	string myLaueGpID;

public:
	spacegroup() { Nsymm = 0; }
	spacegroup(string id, vec3<int> dims) {
		init(id);
		buildMaps(dims);
	}

	// initialize
	//     -- build symmops from string containing spacegroup
	void init(string id) {
		// clear old
		symmMaps.clear();
		symmOps.clear();
		symmInvRot.clear();

		//cout << "spacegroup::spacegroup(" << id << ")\n";
		mySpaceGpID = spacegpID(id);
		getSymmOps(mySpaceGpID, myLaueGpID, symmOps);
		Nsymm = symmOps.size();
		symmInvRot.reserve(Nsymm);

		for (int i=0; i<Nsymm; ++i) {
			symmInvRot.push_back( mat33<double>(symmOps[i].rot).inv() );
		}
	}
	// initialize
	//     -- build symmops from ing containing spacegroup	ID
	void init(int spcgrp) {
		// clear old
		symmMaps.clear();
		symmOps.clear();
		symmInvRot.clear();

		mySpaceGpID = spcgrp;
		getSymmOps(mySpaceGpID, myLaueGpID, symmOps);
		Nsymm = symmOps.size();
		symmInvRot.reserve(Nsymm);

		for (int i=0; i<Nsymm; ++i) {
			symmInvRot.push_back( mat33<double>(symmOps[i].rot).inv() );
		}
	}

	void buildMaps(vec3<int> dims) {
		if (symmMaps.size() == 0 || symmMaps[0].getDims() != dims) {
			if(!symmMaps.empty()){
				symmMaps.clear();
				ptrMapInv.clear();
			}

			symmMaps.reserve(Nsymm);
			for (int symm_idx=0; symm_idx<Nsymm; symm_idx++) {
				map_3d<int> symmMap_i(dims);

				vec3<double> X, SX;
				int idx;
				for (int x=0; x<dims[0]; x++) {
					X[0] = ((double)x)/dims[0];
					for (int y=0; y<dims[1]; y++) {
						X[1] = ((double)y)/dims[1];
						for (int z=0; z<dims[2]; z++) {
							X[2] = ((double)z)/dims[2];
							SX = X.apply(symmOps[symm_idx].rot, symmOps[symm_idx].trans);

							idx = pos_mod( (int)floor(SX[0]*dims[0]+0.5) , dims[0]) * dims[1]*dims[2] +
							      pos_mod( (int)floor(SX[1]*dims[1]+0.5) , dims[1]) * dims[2] +
							      pos_mod( (int)floor(SX[2]*dims[2]+0.5) , dims[2]) ;
							symmMap_i.set(x,y,z, idx);
						}
					}
				}
				symmMaps.push_back(symmMap_i);
			}

			// asymm unit mask
			asymmUnit.alloc(dims); asymmUnit.setAll(0);
			asymmCopy.alloc(dims); asymmCopy.setAll(0);
			ptrMap.alloc(dims);

			int currIdx = 0;
			int N = dims[0]*dims[1]*dims[2];
			ptrMapInv.reserve( N/Nsymm );

			for (int x=dims[0]-1; x>=0; --x)
			for (int y=dims[1]-1; y>=0; --y)
			for (int z=dims[2]-1; z>=0; --z) {
				asymmUnit.set(x,y,z, -1);  // -1 means "current point"

				for (int symm_idx=1; symm_idx<Nsymm; symm_idx++) {   // 1st symmop is identity
					int thisSymmPtr = symmMaps[symm_idx].get(x,y,z);
					if ( asymmUnit.get( thisSymmPtr ) == 1 ) {
						asymmUnit.set(x,y,z, 0);
						asymmCopy.set(x,y,z, symm_idx);
						ptrMap.set   (x,y,z, ptrMap.get( thisSymmPtr ) );
						break;
					}
				}

				if (asymmUnit.get(x,y,z) == -1) {
					asymmUnit.set(x,y,z, 1);
					ptrMapInv.push_back( vec3<int>(x,y,z) );
					ptrMap.set   (x,y,z, currIdx++);
				}
			}
		}
	}

	inline map_3d<int>&         getPtrMap()    { return ptrMap; };
	inline vector< vec3<int> >& getPtrMapInv() { return ptrMapInv; };
	inline int nSymm() { return Nsymm; }
	inline int getId() { return mySpaceGpID; }
	inline string getLaueGpId() { return myLaueGpID; }

	inline mat33<double> getSymmInvRot(int i) {
		return symmInvRot[i];
	}

	inline mat33<double> getSymmRot(int i) {
		return mat33<double>(symmOps[i].rot);
	}

	inline vec3<double> getSymmTrans(int i) {
		return vec3<double>(symmOps[i].trans);
	}

	// get all symmetric copies of a location
	// NOTE: INPUT IN FRACTIONAL COORDS
	void getSymmetric( vec3<double> fcoord , vector< vec3<double> >& Sfcoord) {
		Sfcoord.clear();
		for (int symm_idx=0; symm_idx<Nsymm; symm_idx++) {
			Sfcoord.push_back(fcoord.apply(symmOps[symm_idx].rot, symmOps[symm_idx].trans));
		}
	}

	// chains are represented as a vector of vec3<doubles>
	// function returns a concatenated vector of coordinates
	// NOTE: INPUT IN FRACTIONAL COORDS
	void getSymmetric( vector< vec3<double> > fcoords, vector< vec3<double> > &Sfcoords) {
		int N = fcoords.size();
		Sfcoords.clear(); Sfcoords.reserve(Nsymm*N);

		for (int symm_idx=0; symm_idx<Nsymm; symm_idx++) {
			for (int i=0; i<N; i++) {
				Sfcoords.push_back(
					fcoords[i].apply(symmOps[symm_idx].rot, symmOps[symm_idx].trans)
				);
			}
		}
	}

	// get a specific symmetric copy
	// NOTE: INPUT IN FRACTIONAL COORDS
	void getSymmetric( vector< vec3<double> > fcoords, vector< vec3<double> > &Sfcoords, int copyNum ) {
		int N = fcoords.size();
		Sfcoords.clear(); Sfcoords.reserve(N);

		if (copyNum >= Nsymm) {
			cerr << "spacegroup::getSymmetric(): ERROR!  Illegal symmcopy number.\n";
			return;
		}

		for (int i=0; i<N; i++)
			Sfcoords.push_back( fcoords[i].apply(symmOps[copyNum].rot, symmOps[copyNum].trans) );
	}
	inline void getSymmetric( vec3<double> fcoord, vec3<double> &Sfcoord, int copyNum ) {
		if (copyNum >= Nsymm) {
			cerr << "spacegroup::getSymmetric(): ERROR!  Illegal symmcopy number.\n";
			return;
		}
		Sfcoord = fcoord.apply(symmOps[copyNum].rot, symmOps[copyNum].trans);
	}


	// get all symmetric copies of a _transformation_
	// since we're transforming a transformation, we only nede to apply the
	//      rotation component of the symmetric operator
	// NOTE: INPUT IN FRACTIONAL COORDS
	void getSymmetric(  mat33<double> fRot,
	                    vector< mat33<double> > &sfRot ) {

		for (int symm_idx=0; symm_idx<Nsymm; symm_idx++) {
			sfRot.push_back  ( prod(mat33<double>(symmOps[symm_idx].rot), fRot) );
		}
	}

	template <class T>
	void sumSymmetric( map_3d<T> &M ) {
		if (Nsymm == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims();
		if (symmMaps.size() == 0 || dims != symmMaps[0].getDims()) buildMaps(dims);
		int N = dims[0]*dims[1]*dims[2];

		map_3d<T> Msymm = M;
		for (int symm_idx=1; symm_idx<Nsymm; symm_idx++) {   // 1st symmop is identity
			for (int i=0; i<N; i++) {
				Msymm.set(i, Msymm.get(i) + M.get( symmMaps[symm_idx].get(i) ) );
			}
		}
		M = Msymm;
	}

	template <class T>
	void prodSymmetric( map_3d<T> &M ) {
		if (Nsymm == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims();
		if (symmMaps.size() == 0 || dims != symmMaps[0].getDims()) buildMaps(dims);
		int N = dims[0]*dims[1]*dims[2];

		map_3d<T> Msymm = M;
		for (int symm_idx=1; symm_idx<Nsymm; symm_idx++) {   // 1st symmop is identity
			for (int i=0; i<N; i++) {
				Msymm.set(i, Msymm.get(i) * M.get( symmMaps[symm_idx].get(i) ) );
			}
		}
		M = Msymm;
	}

	template <class T>
	void invProdSymmetric( map_3d<T> &M ) {
		if (Nsymm == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims();
		if (symmMaps.size() == 0 || dims != symmMaps[0].getDims()) buildMaps(dims);
		int N = dims[0]*dims[1]*dims[2];

		M.scaleShift(-Nsymm,1);
		map_3d<T> Msymm = M;
		for (int symm_idx=1; symm_idx<Nsymm; symm_idx++) {   // 1st symmop is identity
			for (int i=0; i<N; i++) {
				Msymm.set(i, Msymm.get(i) * (M.get( symmMaps[symm_idx].get(i) )) );
			}
		}
		M = Msymm;
		M.scaleShift(-1,1);
	}

	vector< map_3d<int> >& getSymmMaps( ) {
		return symmMaps;
	}

	template <class T>
	void maxSymmetric( map_3d<T> &M ) {
		if (Nsymm == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims();
		if (symmMaps.size() == 0 || dims != symmMaps[0].getDims()) buildMaps(dims);
		int N = dims[0]*dims[1]*dims[2];

		T maxM_i;
		map_3d<T> Morig = M;
		for (int i=0; i<N; i++) {
			maxM_i = Morig.get(i);
			for (int symm_idx=1; symm_idx<Nsymm; symm_idx++)   // 1st symmop is identity
				maxM_i = max(maxM_i , Morig.get( symmMaps[symm_idx].get(i) ) );
			M.set(i, maxM_i);
		}
	}

	template <class T>
	void maxSymmetric( map_3d<T> &M , map_3d<int> &ptr , map_3d<int> &symop) {
		if (Nsymm == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims();
		if (symmMaps.size() == 0 || dims != symmMaps[0].getDims()) buildMaps(dims);
		int N = dims[0]*dims[1]*dims[2];

		T maxM_i, M_i;
		int argmaxM_i, arg_i, symmaxM_i;

		map_3d<T> Morig = M;
		for (int i=0; i<N; i++) {
			maxM_i = Morig.get(i);
			argmaxM_i = i;
			symmaxM_i = 0;

			for (int symm_idx=1; symm_idx<Nsymm; symm_idx++) {
				arg_i = symmMaps[symm_idx].get(i);
				M_i = Morig.get( arg_i );
				if (M_i > maxM_i) {
					maxM_i = M_i;
					argmaxM_i = arg_i;
					symmaxM_i = symm_idx;
				}
			}

			M.set(i, maxM_i);
			ptr.set(i,argmaxM_i);
			symop.set(i,symmaxM_i);
		}
	}

	template <class T>
	void maskAsymmetricUnit( map_3d<T> &M , T maskVal) {
		if (Nsymm == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims(), dimsB = asymmUnit.getDims();
		if (dims[0] != dimsB[0] || dims[1] != dimsB[1] || dims[2] != dimsB[2]) {
			cerr << "spacegroup::maskAsymmetricUnit(): map " << dims.to_string() << " and symmMap "
											                 << dimsB.to_string() << " dimensions do not match!\n";
			return;
		}

		int N = dims[0]*dims[1]*dims[2];
		for (int i=0; i<N; i++) {
			if ( asymmUnit.get( i ) <= 0 ) {
				M.set(i,maskVal);
			}
		}
	}

	map_3d<short>& getAsymmUnit() { return asymmUnit; }
	map_3d<short>& getAsymmCopy() { return asymmCopy; }
};



/////////////////////////////////////////
//
// crystal parameters
//
class Crystal {
private:
	spacegroup           symmGp;

public:
	string               symmGpId, crystline;
	double               A,B,C,alpha,beta,gamma;          // real space unit cell dims
	double               rA,rB,rC,crAlpha,crBeta,crGamma; // reciprocal space unit cell dims
	double               V, rV;                           // unit cell volume
	mat33<double>        f2c,c2f;


	Crystal() {
		setCrystal(1.0,1.0,1.0, 90.0,90.0,90.0);
		symmGpId = "P 1";
	}

	Crystal(string crystfile) {
		ifstream crystin(crystfile.c_str());
		if (!crystin.is_open()) {
			cerr << "ERROR: Unable to open crystal parameters '" << crystfile << "'\n";
			throw 0;
		}

		// first line contains crystal info
		getline(crystin,crystline);
		init(crystline);
		crystin.close();
	}

	// construct from parameters
	Crystal(double A, double B, double C, double alpha, double beta, double gamma) {
		setCrystal(A, B, C, alpha, beta, gamma);
		symmGpId = "P 1";
	}

	Crystal(double A, double B, double C, double alpha, double beta, double gamma, string id) {
		setCrystal(A, B, C, alpha, beta, gamma);
		// init symmetric operators
		symmGpId = id; trim(symmGpId);
		symmGp.init(symmGpId);
	}
	Crystal(double A, double B, double C, double alpha, double beta, double gamma, int spcgrpid) {
		setCrystal(A, B, C, alpha, beta, gamma);
		// init symmetric operators
		symmGp.init(spcgrpid);
	}
	void init(string crystline) {
		//setCrystal( atof(crystline.substr( 6,8)), atof(crystline.substr(15,8)), atof(crystline.substr(24,8)),
		//            atof(crystline.substr(33,6)), atof(crystline.substr(40,6)), atof(crystline.substr(47,6)));

		// more robust parsing
		//this->crystline = crystline;
		vector<string> crystTokens;
		Tokenize(crystline, crystTokens, " ,");

		int startTokenIdx = 0;
		if (crystTokens[0] == "CRYST1") startTokenIdx++;

		setCrystal( atof(crystTokens[startTokenIdx+0]),
		            atof(crystTokens[startTokenIdx+1]),
		            atof(crystTokens[startTokenIdx+2]),
		            atof(crystTokens[startTokenIdx+3]),
		            atof(crystTokens[startTokenIdx+4]),
		            atof(crystTokens[startTokenIdx+5]) );

		symmGpId = crystTokens[startTokenIdx+6];
		for (int i=startTokenIdx+6; i<crystTokens.size(); ++i) {
			symmGpId = symmGpId+" "+crystTokens[i];
		}

		// init symmetric operators
		symmGpId = crystline.substr(55,10); trim(symmGpId);
		symmGp.init(symmGpId);
		stringstream line;
		line.setf(ios::fixed,ios::floatfield);
		line << "CRYST1" << setw(9) << setprecision(3) << this->A;
		line << setw(9) << setprecision(3) << this->B;
		line << setw(9) << setprecision(3) << this->C;
		line << setw(7) << setprecision(2) << this->alpha;
		line << setw(7) << setprecision(2) << this->beta;
		line << setw(7) << setprecision(2) << this->gamma;
		line << " " << setiosflags(ios::left) << symmGpId;
		this->crystline = line.str();
	}

	void setCrystal(double A, double B, double C, double alpha, double beta, double gamma) {
		this->A = A;
		this->B = B;
		this->C = C;
		this->alpha = alpha;
		this->beta = beta;
		this->gamma = gamma;

		// init transformation matrices
		double ca = cos(d2r(alpha)), cb = cos(d2r(beta)), cg = cos(d2r(gamma));
		double sa = sin(d2r(alpha)), sb = sin(d2r(beta)), sg = sin(d2r(gamma));
		this->f2c = mat33<double>( A  , B * cg, C * cb,
							 0.0, B * sg, C * (ca - cb*cg) / sg,
							 0.0, 0.0   , C * sb * sqrt(1.0 - square((cb*cg - ca)/(sb*sg))) );
		this->c2f = f2c.inv();
		this->V = A*B*C* sqrt(1-square(ca)-square(cb)-square(cg)+2*ca*cb*cg);

		// reciprocal space cell dimensions
		this->rA = B*C*sa/V;
		this->rB = A*C*sb/V;
		this->rC = A*B*sg/V;
		this->crAlpha = cos(  asin( min( max( V/(A*B*C*sb*sg) , -1.0) , 1.0) )  );
		this->crBeta  = cos(  asin( min( max( V/(A*B*C*sa*sg) , -1.0) , 1.0) )  );
		this->crGamma = cos(  asin( min( max( V/(A*B*C*sa*sb) , -1.0) , 1.0) )  );
		this->rV = 1/V;

	}

	// num symmetric
	int nSymm() { return symmGp.nSymm(); }

	inline mat33<double> getSymmInvRot(int i) {
		return symmGp.getSymmInvRot(i);
	}
	inline mat33<double> getSymmRot(int i) {
		return symmGp.getSymmRot(i);
	}
	inline vec3<double> getSymmTrans(int i) {
		return symmGp.getSymmTrans(i);
	}

	// get symmetric, asymmetric copies
	void getSymmetric(vec3<double> fX , vector< vec3<double> >& SfX) {
		symmGp.getSymmetric(fX,SfX);
	}
	void getSymmetric(vector< vec3<double> > fX , vector< vec3<double> >& SfX, int N) {
		symmGp.getSymmetric(fX,SfX, N);
	}
	void getSymmetric(vec3<double> fX , vec3<double> &SfX, int N) {
		symmGp.getSymmetric(fX,SfX, N);
	}

	// sum symmetric
	template <class T> void sumSymmetric(map_3d<T> &M) { symmGp.sumSymmetric(M); }
	template <class T> void prodSymmetric(map_3d<T> &M) { symmGp.prodSymmetric(M); }
	template <class T> void invProdSymmetric(map_3d<T> &M) { symmGp.invProdSymmetric(M); }
	template <class T> void maxSymmetric(map_3d<T> &M) { symmGp.maxSymmetric(M); }
	template <class T> void maxSymmetric(map_3d<T> &M, map_3d<int> &ptr, map_3d<int> &sym) { symmGp.maxSymmetric(M, ptr, sym); }

	template <class T> void minSymmetric(map_3d<T> &M) {
		M.scale(-1);
		symmGp.maxSymmetric(M);
		M.scale(-1);
	}
	template <class T> void minSymmetric(map_3d<T> &M, map_3d<int> &ptr, map_3d<int> &sym) {
		M.scale(-1);
		symmGp.maxSymmetric(M, ptr, sym);
		M.scale(-1);
	}

	// raw symmmap access
	vector< map_3d<int> >& getSymmMaps( ) { return symmGp.getSymmMaps(); }
	void buildMaps( vec3<int> &dims ) { return symmGp.buildMaps(dims); }

	// get map dimensions that are SYMMGROUP-APPROPRIATE
	//     use same algo as setgrd.f from CCP4
	vec3<int> getDims(double approxSpacing) {
		int NX=2,NY=2,NZ=2;

		string laueGpId=symmGp.getLaueGpId();

		if (laueGpId == "PG1"   || laueGpId == "PG1bar")    { NX = 2; NY = 2; NZ = 2; }
		if (laueGpId == "PG2")                              { NX = 2; NY = 4; NZ = 2; }
		if (laueGpId == "PG2/m" || laueGpId == "PG2c")      { NX = 2; NY = 8; NZ = 4; }
		if (laueGpId == "PG222" || laueGpId == "PGmmm")     { NX = 4; NY = 4; NZ = 4; }
		if (laueGpId == "PG4"   || laueGpId == "PG4/m")     { NX = 4; NY = 4; NZ = 8; }
		if (laueGpId == "PG422" || laueGpId == "PG4/mmm")   { NX = 4; NY = 4; NZ = 8; }
		if (laueGpId == "PG3"   || laueGpId == "PG3bar")    { NX = 6; NY = 6; NZ = 6; }
		if (laueGpId == "PG312" || laueGpId == "PG312")     { NX = 6; NY = 6; NZ = 6; }
		if (laueGpId == "PG321" || laueGpId == "PG312")     { NX = 6; NY = 6; NZ = 6; }
		if (laueGpId == "PG6"   || laueGpId == "PG6/m")     { NX = 6; NY = 6; NZ = 12; }
		if (laueGpId == "PG622" || laueGpId == "PG6/mmm")   { NX = 6; NY = 6; NZ = 12; }
		if (laueGpId == "PG23"  || laueGpId == "PGm3bar")   { NX = 4; NY = 4; NZ = 4; }
		if (laueGpId == "PG432" || laueGpId == "PGm3barm")  { NX = 8; NY = 8; NZ = 8; }

		vec3<int> dims( findSampling(A,NX),findSampling(B,NY),findSampling(C,NZ) );

		return dims;
	}

	////////////////////
	///
	/// asymm unit operators
	template <class T>
	void maskAsymmetricUnit(map_3d<T> &M, T maskVal = 0.0) {
		symmGp.maskAsymmetricUnit(M, maskVal);
	}
	void maskSymmetricCollisions(map_3d<double> &M, double mindist) {
		if (symmGp.nSymm() == 0) return;  // uninitialized spacegroup

		vec3<int> dims = M.getDims();
		double mindist2 = mindist * mindist;

		M.setAll(1.0);

		for (int i=0; i<dims[0]; i++)
		for (int j=0; j<dims[1]; j++)
		for (int k=0; k<dims[2]; k++) {
			vec3<double>  fx( ((double)i)/dims[0], ((double)j)/dims[1], ((double)k)/dims[2] ),
			             sfx , delfx;

			for (int symm_idx=1; symm_idx<symmGp.nSymm(); symm_idx++) {   // 1st symmop is identity
				getSymmetric( fx, sfx, symm_idx );
				delfx = fx - sfx; delfx.minmod1();

				if ( delfx.apply(f2c).norm2() <= mindist2 ) {
					M.set(i,j,k,0.0);
					break;
				}
			}
		}
	}
	inline map_3d<int>&         getPtrMap()    { return symmGp.getPtrMap(); }
	inline vector< vec3<int> >& getPtrMapInv() { return symmGp.getPtrMapInv(); }
	map_3d<short>& getAsymmUnit() { return symmGp.getAsymmUnit(); }
	map_3d<short>& getAsymmCopy() { return symmGp.getAsymmCopy(); }



	// get S
	double S(int h, int k, int l) {
		return ( square(h*rA) + square(k*rB) + square(l*rC)
		           + 2*h*k*rA*rB*crGamma
		           + 2*h*l*rA*rC*crBeta
		           + 2*k*l*rB*rC*crAlpha );
	}

	string to_string() {
		std::ostringstream oss;
		oss << A << " x " << B << " x " << C << " @ " << alpha << " , " << beta << " , " << gamma ;
		return oss.str();
	}

};

////////////////////////////
//
// read an CCP4 dens map
//
////////////////////////////
bool readCCP4Map(map_3d<double>& map, Crystal& crystInfo, string mapfile) {
	ifstream mapin(mapfile.c_str(), ios::in|ios::binary);

	if (!mapin.is_open())  {
		cerr << "ERROR: Unable to open CCP4 file '" << mapfile << "'\n";
		return false;
	}
	//Verify this is a CCP4 file
	char* buffer;
	buffer = new char[4];

	mapin.seekg(208, ios::beg);
	mapin.read(buffer,3);
	buffer[3] = '\0';
	if(!mapin || strcmp(buffer,"MAP") != 0){
		cerr << "ERROR:  'MAP' string missing, not a valid CCP4 file." << endl;
		return false;
	}
	delete[] buffer;

	mapin.seekg(0, ios::beg);

	vec3<int> dims;
	int extent[3], mapmode, origin[3],gridsize[3],crs2xyz[3], spcgrpID;
	float cellDims[3], cellAngles[3], densStats[3];
	int byteOrder = 1;
	//Read number of columns,rows,sections
	ReadWord(mapin,(char*)(&extent[0]),byteOrder);
	ReadWord(mapin,(char*)(&extent[1]),byteOrder);
	ReadWord(mapin,(char*)(&extent[2]),byteOrder);

	//Read CCP4 map mode for storing values
	ReadWord(mapin,(char*)(&mapmode),byteOrder);
	if(mapmode != 2){
		mapin.seekg(0,ios::beg);
		byteOrder = -1;
		//Reread number of columns,rows,sections
		ReadWord(mapin,(char*)(&extent[0]),byteOrder);
		ReadWord(mapin,(char*)(&extent[1]),byteOrder);
		ReadWord(mapin,(char*)(&extent[2]),byteOrder);

		//Reread CCP4 map mode for storing values
		ReadWord(mapin,(char*)(&mapmode),byteOrder);
		if(mapmode !=2){
			cerr << "ERROR:  Non-real (32-bit float) data types are unsupported." << endl;
			return false;
		}
	}
	//Read origin
	ReadWord(mapin,(char*)(&origin[0]),byteOrder);
	ReadWord(mapin,(char*)(&origin[1]),byteOrder);
	ReadWord(mapin,(char*)(&origin[2]),byteOrder);
	//Read grid size
	ReadWord(mapin,(char*)(&gridsize[0]),byteOrder);
	ReadWord(mapin,(char*)(&gridsize[1]),byteOrder);
	ReadWord(mapin,(char*)(&gridsize[2]),byteOrder);
	//Read cell dimenions and angles
	ReadWord(mapin,(char*)(&cellDims[0]),byteOrder);
	ReadWord(mapin,(char*)(&cellDims[1]),byteOrder);
	ReadWord(mapin,(char*)(&cellDims[2]),byteOrder);
	ReadWord(mapin,(char*)(&cellAngles[0]),byteOrder);
	ReadWord(mapin,(char*)(&cellAngles[1]),byteOrder);
	ReadWord(mapin,(char*)(&cellAngles[2]),byteOrder);
	//Read axis order and density statistics
	ReadWord(mapin,(char*)(&crs2xyz[0]),byteOrder);
	ReadWord(mapin,(char*)(&crs2xyz[1]),byteOrder);
	ReadWord(mapin,(char*)(&crs2xyz[2]),byteOrder);
	mapin.seekg(12,ios::cur);
	//Read Space Group ID Number
	ReadWord(mapin,(char*)(&spcgrpID),byteOrder);
	if(!mapin){
		cerr << "ERROR:  Improperly formatted line" << endl;
		return false;
	}
	//Save crystal information
	crystInfo = Crystal(cellDims[0], cellDims[1], cellDims[2], cellAngles[0],cellAngles[1],cellAngles[2], spcgrpID );

	// check extent and grid interval counts
	if (gridsize[0] == 0 && extent[0] > 0) {
		gridsize[0] = extent[0] - 1;
		cerr << "WARNING: Fixed X interval count" << endl;
	}
	if (gridsize[1] == 0 && extent[1] > 0) {
		gridsize[1] = extent[1] - 1;
		cerr << "WARNING: Fixed Y interval count" << endl;
	}
	if (gridsize[2] == 0 && extent[2] > 0) {
		gridsize[2] = extent[2] - 1;
		cerr << "WARNING: Fixed Z interval count" << endl;
	}
	// Mapping between CCP4 column, row, section and VMD x, y, z.
	if (crs2xyz[0] == 0 && crs2xyz[1] == 0 && crs2xyz[2] == 0) {
		cerr << "WARNING: All crs2xyz records are zero." << endl;
		cerr << "WARNING: Setting crs2xyz to 1, 2, 3" << endl;
		crs2xyz[0] = 1;
		crs2xyz[1] = 2;
		crs2xyz[2] = 3;
	}

	int xyzOrder[3];
	xyzOrder[crs2xyz[0]-1]=0;
	xyzOrder[crs2xyz[1]-1]=1;
	xyzOrder[crs2xyz[2]-1]=2;


	//Read Data
	mapin.seekg(-4*extent[0]*extent[1]*extent[2], ios::end);

	map.alloc(extent[xyzOrder[0]],extent[xyzOrder[1]],extent[xyzOrder[2]]);
	int CRS[3];
	for(CRS[2] = 0; CRS[2] < extent[2];CRS[2]++){
		for(CRS[1] = 0; CRS[1] < extent[1]; CRS[1]++){
			for(CRS[0] = 0; CRS[0] < extent[0]; CRS[0]++){
				float value;
				if(!mapin){
					cerr << "Error Reading Data, not enough data" << endl;
					return false;
				}
				ReadWord(mapin,(char*)&value, byteOrder);
				map.set(CRS[xyzOrder[0]],CRS[xyzOrder[1]],CRS[xyzOrder[2]],(double)value);
			}
		}
	}
	mapin.close();

	//extend to full unit cell
	map_3d<double> newDensity(gridsize[0],gridsize[1],gridsize[2], map.getMin());

	// if it already covers unit cell no worries
	vec3<int> newGrid (gridsize[0],gridsize[1],gridsize[2]);
	crystInfo.buildMaps( newGrid );
	gridsize[0] = newGrid[0];gridsize[1] = newGrid[1];gridsize[2] = newGrid[2];

	int newX,newY,newZ;
	for (int x=0; x<map.getDims(0); ++x) {
		newX = pos_mod( x+origin[0] , gridsize[0] );
		for (int y=0; y<map.getDims(1); ++y) {
			newY = pos_mod( y+origin[1] , gridsize[1] );
			for (int z=0; z<map.getDims(2); ++z) {
				newZ = pos_mod( z+origin[2] , gridsize[2] );
				newDensity.set( newX,newY,newZ, map.get(x,y,z));
			}
		}
	}

	//save results
	crystInfo.maxSymmetric( newDensity );
	map = newDensity;
	//Warn if too much of the unit cell is zero
	int zeros = 100-(map.numNonZero()*100/(map.getDimsProd()));
	if(zeros>20){
		cerr << "WARNING: Electron Density Map is " << zeros <<"% zeroes, please ensure the density values cover the entire unit cell" << endl;
	}
	return true;
}
////////////////////////////
//
// read an FS4 dens map
//
////////////////////////////
bool readFS4Map(map_3d<double>& map, string mapfile) {
	ifstream mapin(mapfile.c_str(), ios::in|ios::binary);

	if (!mapin.is_open())  {
		cerr << "ERROR: Unable to open match map '" << mapfile << "'\n";
		return false;
	}

	int readBuf, jump=4, ordering, byteOrder=1;
	vec3<int> dims;
	char map2fs_test[6];

	// was this map generated by map2fs??  if so, skip extra header
	mapin.seekg(27,ios_base::beg);
	mapin.read(map2fs_test, 6);
	if (!strncmp(map2fs_test, "map2fs", 6))
		jump += 44;
	mapin.seekg(0,ios_base::beg);  // back to beginning of file

	// read the first word in the file, which describes the size of
	//   the header block.  Determine Endianness. Skip ahead.
	unsigned char byte_in[4];
	mapin.read((char*)byte_in, 4);

	if ((byte_in[0]) + (byte_in[1] << 8) > (byte_in[3]) + (byte_in[2] << 8)) {
		byteOrder = 1;
		readBuf = (byte_in[0]) + (byte_in[1] << 8);
	} else {
		byteOrder = -1;
		readBuf = (byte_in[3]) + (byte_in[2] << 8);
	}
	mapin.seekg(readBuf+jump, ios_base::cur);

	// Next block is the map stats block ... read the first 5 words
	//     word 1    size of symmetry block
	//      "   2    nrow
	//      "   3    ncol
	//      "   4    nplane
	//      "   5    norn = ordering of axes: 0 (XZY), 1 (YZX), or 2 (XYZ)
	ReadWord(mapin, (char*)&readBuf , byteOrder);
	ReadWord(mapin, (char*)&dims[0] , byteOrder);
	ReadWord(mapin, (char*)&dims[1] , byteOrder);
	ReadWord(mapin, (char*)&dims[2] , byteOrder);
	mapin.seekg(4, ios_base::cur);
	ReadWord(mapin, (char*)&ordering , byteOrder);

	// read and store in >XYZ< ordering
	int _x, _y, _z;
	switch(ordering) {
		case 0:  _x = 2; _y = 0; _z = 1; break;
		case 1:  _x = 0; _y = 1; _z = 2; break;
		default: _x = 1; _y = 0; _z = 2;
	}

	map.alloc(dims[_y],dims[_x],dims[_z]);

	//cout << "Read match map '" << mapfile << "' " << map.getDims().to_string() << "\n";

	// Now begin reading the data
	// First skip ahead to the data block
	mapin.seekg(readBuf-16, ios_base::cur);

	vec3<int> X;
	for (X[2]=0; X[2]<dims[2]; X[2]++) {
		for (X[1]=0; X[1]<dims[1]; X[1]++) {
			mapin.seekg(4, ios_base::cur);
			for (X[0]=0; X[0]<dims[0]; X[0]++) {
				ReadWord(mapin, (char*)&readBuf, byteOrder);
				map.set(X[_y],X[_x],X[_z],(double)readBuf);
			}
			mapin.seekg(4, ios_base::cur);
		}
	}

	mapin.close();
	//Warn if too much of the unit cell is zero
	int zeros = 100-(map.numNonZero()*100/(map.getDimsProd()));
	if(zeros>20){
		cerr << "WARNING: Electron Density Map is " << zeros <<"% zeroes, please ensure the density values cover the entire unit cell" << endl;
	}
	return true;
}
//Decide what format the input map is, and read the map and crystal information
bool readDensityMap(map_3d<double>& map, string mapfile, Crystal& crystInfo, string crystfile){
	ifstream mapin(mapfile.c_str(), ios::in|ios::binary);

	if (!mapin.is_open())  {
		cerr << "ERROR: Unable to open map file '" << mapfile << "'\n";
		return false;
	}
	char * buffer;
	buffer = new char[4];
	mapin.seekg(208, ios::beg);
	mapin.read(buffer,3);
	buffer[3] = '\0';
	if(mapin && strcmp(buffer,"MAP") == 0){
		mapin.close();
		return readCCP4Map(map,crystInfo,mapfile);
	} else {
		try{
			crystInfo = Crystal(crystfile);
		} catch (int e){
			cerr << "Error:Crystal file not valid, required for FS4 density maps" << endl;
			return false;
		}
		mapin.close();
		return readFS4Map(map,mapfile);
	}
}
#endif
