// //////////////////////////////////////////////////////////////////////////
//
//     nmer_lookup.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// ////////////////////////////////////////////////////////////////////////////

#include "nmer_db.h"

#include <stdio.h>
#include <stdlib.h>
#include <string>
#include <vector>
#include <iterator>
#include "protein.h"

#define NMER 5

///////////////////////////////////////////////////////
//
// GLOBALS
//
static string pdb_filedir; //  = PDB_FILEDIR;
static string pdb_database; // = PDB_DATABASE;

//
///////////////////////////////////////////////////////


///////////////////////////////////////////////////////
//
//
//  query environment variable 'ACMI_NMERDB' to find nmer db directory
void get_DB_paths(string &database, string &filedir) {
	char* envvar = getenv("ACMI_NMERDB");

	if (envvar == NULL) {
		database = "../nmerdb/nmers.db";
		filedir  = "../nmerdb/fraglib/";
	} else {
		database = string(envvar) + "/nmers.db";
		filedir  = string(envvar) + "/fraglib/";
	}
}


// find the min entry in a SYMMETRIC distance matrix
inline double distMatrixMin(vector< vector<double> > D , int &minI, int &minJ) {
	minI = minJ = -1;
	double minVal = INFINITY;

	for(int i=0; i<D.size(); i++) {
		for (int j=i+1; j<D.size(); j++) {
			if (D[i][j] < minVal) {
				minVal = D[i][j];
				minI = i;
				minJ = j;
			}
		}
	}

	return minVal;
}

// join the second vector to the end of the first, then clear the second
template <class T>
inline void appendAndClear(vector<T> &X, vector<T> &Y) {
	int N=Y.size();
	for (int i=0; i<N; i++)
		X.push_back(Y[i]);
	Y.clear();
}
void getRandomFrag(Nmer& frag,double radius){
	//static hash_map< int,  vector< nmerDBEntry > >   DB_nmer;
	static map< int,  vector< nmerDBEntry > >   DB_nmer;
	static vector< string >   DB_filename;
	static vector<int> Keys;
	//for()
	// load db the first time this funcion is called
	if (DB_filename.size() == 0) {
		get_DB_paths(pdb_database, pdb_filedir);
		chem_db_initialize();
		DB_load(pdb_database, DB_nmer , DB_filename);
	}
	int rand_idx = int(DB_nmer.size()*rand()/(RAND_MAX*1.0));
	map< int,  vector< nmerDBEntry > >::iterator item = DB_nmer.begin();
	std::advance(item,rand_idx);
	vector< nmerDBEntry > &nmerlist = item->second;
	rand_idx = int(nmerlist.size()*rand()/(RAND_MAX*1.0));
	string pdb_file = pdb_filedir + "/" + DB_filename[ nmerlist[rand_idx].fileId ];

	if ( exists(pdb_file) ) {

		frag.readFile(pdb_file, nmerlist[rand_idx].chainId, nmerlist[rand_idx].resId, radius);
	}
}

//
//  recursively lookup 5-mers at a PAM distance 'cost' from
//
void pam(vector<Nmer> &fragList,
         const char* seqL,
		 const char* seqR,
         int costRemain,
		 int costUsed,
		 int positionalCost[],
		 double radius) {
	//static hash_map< int,  vector< nmerDBEntry > >   DB_nmer;
	static map< int,  vector< nmerDBEntry > >   DB_nmer;
	static vector< string >   DB_filename;

	// load db the first time this funcion is called
	if (DB_filename.size() == 0) {
		get_DB_paths(pdb_database, pdb_filedir);
		chem_db_initialize();
		DB_load(pdb_database, DB_nmer , DB_filename);
	}

	char newSeqR[NMER];
	int pamIdx, i, addedCost;
	Nmer toInsert(NMER);

	if ((int)seqL[0] == 0) {
		if (costRemain == 0) {  // exact cost matches only

			// DO DB LOOKUP of 'seqR'
			//
			// SELECT pdb_file,chain,res_id FROM 5mers WHERE 5mers.nmer=$seqR
			int nmerhash = chem_db_hashcode(seqR);
			if (DB_nmer.find(nmerhash) == DB_nmer.end())
				return;  // no such 5-mer

			vector< nmerDBEntry > &nmerlist = DB_nmer[nmerhash];
			for (int i=0; i<nmerlist.size(); ++i) {
				string pdb_file = pdb_filedir + "/" + DB_filename[ nmerlist[i].fileId ];

				if ( exists(pdb_file) ) {

					toInsert.readFile(pdb_file, nmerlist[i].chainId, nmerlist[i].resId, radius);
					if ( toInsert.legalStructure(seqR) ) {
						//cerr << fragList.size() << ": " << pdb_file << " , " << nmerlist[i].chainId << " , " << nmerlist[i].resId << endl;
						fragList.push_back( toInsert );
					}
				}
			}


		}
	} else {
		// don't allow '_'s OR the center position to be substituted
		if (seqL[0] == '_' || positionalCost[0] >= 100) {
			sprintf(newSeqR, "%s%c", seqR, seqL[0]);
			pam(fragList, &seqL[1], newSeqR, costRemain, costUsed, &positionalCost[1], radius);
		} else {
			pamIdx = AA2ID[ (unsigned int)seqL[0] ];
			for (i=0; i<23; i++) {
				addedCost = positionalCost[0] * PAM250[pamIdx][i];
				if (addedCost <= costRemain) {
					sprintf(newSeqR, "%s%c", seqR,ID2AA[i]);
					pam(fragList, &seqL[1], newSeqR, costRemain-addedCost,
					    costUsed+addedCost, &positionalCost[1], radius);
				}
			}
		}
	}
}

//
// nmer database lookup
//
void nmer_lookup(
			string mer,
			int MIN_EXAMPLES,
			vector<Nmer> &frags,
			int POSITIONAL_COST[],
			int MAX_PAM_DISTANCE,
			double radius = 0.0) {
	frags.clear();

	int maxCost = 0;
	while (maxCost <= MAX_PAM_DISTANCE && frags.size() < MIN_EXAMPLES) {
		pam(frags,mer.c_str(),"",maxCost++,0,POSITIONAL_COST, radius);
	}
}



//
// nmer lookup in PDB file
//
//    TODO: return 5-mers w/ close PAM scores (now only srches for exact)
void nmer_lookupInFile(
             string   mer,
			 string   pdbfile,
             CAtrace &input,
			 vector<Nmer> &frags,
			 double radius = 0.0) {
	int idx=0, resId;
	char resType,chnId;
	for (int i=0; i<input.nres(); ++i) {
		if (mer == input.serialGet5mer(i)) {
			input.serialGet(i, resId, resType, chnId);
			frags.push_back( Nmer(pdbfile,chnId,resId,radius) );
			//cout << i << " " << resId << endl;
		}
	}
}
