// //////////////////////////////////////////////////////////////////////////
//
//     soft_wrapper.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


extern "C" {
#include "fftw3.h"
#include "complex.h"
#include "csecond.h"
#include "so3_correlate_fftw.h"
#include "soft_fftw_pc.h"
#include "soft_fftw.h"
#include "makeWigner.h"
#include "makeweights.h"
#include "FST_semi_memo.h"
#include "cospmls.h"
#include "primitive_FST.h"
#include "seminaive.h"
#include "rotate_so3.h"
}


// globals -- share memory for sharm xforms
fftw_plan  dctPlan, idctPlan, fftPlan, ifftPlan, p1;
complex<double> *workspace1, *workspace2, *so3Coef;
double *workspace3, *seminaive_naive_tablespace, *trans_seminaive_naive_tablespace, *weights, *wignersTrans;
double *rdata, *idata;
double **seminaive_naive_table, **trans_seminaive_naive_table;
map_3d< complex<double> > so3Sig, so3SigSum;

// globals
int        B, Bout;
double     Rstep;



// 3d cubic spline coeff
struct cubicSplineCoeff3d{
 	int   idx[3];
	float wt[3][4];
} ;

// 3d linear coeff
struct linearCoeff3d{
 	vec3<int> ptLower;
	float     wts[8];
} ;

//
// could do this as a table lookup to save time
inline double factorial(int k) {
	double retval = 1.0;
	for (int i=2; i <= k; i++)
		retval *= (double)i;
	return retval;
}


//
// compute  Wigner-3j
//
double Wigner3j(int j1, int j2, int j3, int m1, int m2, int m3) {
	// error checking
	if (j3 > j1 + j2 || j3 < abs(j1 - j2)) {
		cerr << "ERROR: (Wigner3j) j3 is out of bounds.\n";
		return 0;
	}

	if (abs(m1) > j1) {
		cerr << "ERROR: (Wigner3j) m1 is out of bounds.\n";
		return 0;
	}

	if (abs(m2) > j2) {
		cerr << "ERROR: (Wigner3j) m2 is out of bounds.\n";
		return 0;
	}

	if (abs(m3) > j3) {
		cerr << "ERROR: (Wigner3j) m3 is out of bounds.\n";
		return 0;
	}

	int t1 = j2 - m1 - j3;
	int t2 = j1 + m2 - j3;
	int t3 = j1 + j2 - j3;
	int t4 = j1 - m1;
	int t5 = j2 + m2;

	int tmin = max( 0, max( t1, t2 ) );
	int tmax = min( t3, min( t4, t5 ) );

	double wigner = 0;

	for (int t = tmin; t <=tmax; ++t) {
		wigner += ((t%2)==0 ? 1.0 : -1.0) / (  factorial(t) * factorial(t-t1) * factorial(t-t2)
		                                   * factorial(t3-t) * factorial(t4-t) * factorial(t5-t) );
	}

	wigner *= (((j1-j2-m3) % 2)==0 ? 1.0:-1.0)
	            * sqrt(  factorial(j1+j2-j3) * factorial(j1-j2+j3) * factorial(-j1+j2+j3) / factorial(j1+j2+j3+1)
	                   * factorial(j1+m1) * factorial(j1-m1) * factorial(j2+m2) * factorial(j2-m2) * factorial(j3+m3) * factorial(j3-m3) );

	return wigner;
}

//
// compute Clebsh-Gordon coefficients
//
double ClebschGordan(int j1, int j2, int j, int m1, int m2, int m) {
	// error checking
	if (m1 + m2 != m) {
		cerr << "ERROR: (ClebschGordan) m1+m2 must equal m.\n";
		return 0;
	}
	if (j > j1 + j2 || j < abs(j1 - j2)) {
		//cerr << "ERROR: (ClebschGordan) j is out of bounds.\n";
		return 0;
	}
	if (abs(m1) > j1) {
		//cerr << "ERROR: (ClebschGordan) m1 is out of bounds.\n";
		return 0;
	}
	if (abs(m2) > j2) {
		//cerr << "ERROR: (ClebschGordan) m2 is out of bounds.\n";
		return 0;
	}
	if (abs(m) > j) {
		//cerr << "ERROR: (ClebschGordan) m is out of bounds.\n";
		return 0;
	}

	double sign = (((j1-j2+m) % 2) == 0)? 1.0 : -1.0;
	double cg = sign * sqrt(2.0*j + 1.0) * Wigner3j(j1,j2,j,m1,m2,-m);

	return cg;
}

// initialize SOFT
//     assumes globals B, Bout initialized
void SOFT_initialize() {
	workspace1 = new complex<double>[8*Bout*Bout*Bout];
	workspace2 = new complex<double>[14*B*B + 48*B];
	workspace3 = new double[24*B + 2*B*B];

	rdata = new double[4*B*B];
	idata = new double[4*B*B];
	for (int i=0; i<4*B*B; i++) idata[i] = 0;

	so3Coef = new complex<double>[ (4*Bout*Bout*Bout-Bout)/3 ] ;
	seminaive_naive_tablespace
	                = new double[Reduced_Naive_TableSize(B,B)+Reduced_SpharmonicTableSize(B,B)];
	trans_seminaive_naive_tablespace
	                = new double[Reduced_Naive_TableSize(B,B)+Reduced_SpharmonicTableSize(B,B)];

	weights = new double[4*B];
	makeweights( B, weights );

	so3Sig.alloc(2*Bout,2*Bout,2*Bout);
	so3SigSum.alloc(2*Bout,2*Bout,2*Bout);

	////////////////
	// set up fftw plans
	//
	fftw_iodim sht_dims[1], sht_howmany_dims[1];

	//  (1a) sht plan
	dctPlan = fftw_plan_r2r_1d( 2*B, weights, rdata, FFTW_REDFT10, FFTW_ESTIMATE ) ;
	sht_dims[0].n = 2*B ; sht_dims[0].is = 1 ; sht_dims[0].os = 2*B ;
	sht_howmany_dims[0].n = 2*B ; sht_howmany_dims[0].is = 2*B ; sht_howmany_dims[0].os = 1 ;
	fftPlan = fftw_plan_guru_split_dft( 1, sht_dims, 1, sht_howmany_dims, rdata, idata,
	                                              (double*) (&workspace2[0]), (double*) (&workspace2[2*B*B]),
	                                              FFTW_ESTIMATE );

	//  (1b) isht plan
	idctPlan = fftw_plan_r2r_1d( 2*B, weights, rdata, FFTW_REDFT01, FFTW_ESTIMATE ) ;
	sht_dims[0].n = 2*B ; sht_dims[0].is = 2*B ; sht_dims[0].os = 1 ;
	sht_howmany_dims[0].n = 2*B ; sht_howmany_dims[0].is = 1 ; sht_howmany_dims[0].os = 2*B ;
	ifftPlan = fftw_plan_guru_split_dft( 1, sht_dims, 1, sht_howmany_dims, rdata, idata,
						(double*) (&workspace2[0]), (double*) (&workspace2[2*B*B]),
	                                              FFTW_ESTIMATE );

	//
	// (2) SO3 inv transform fftw plan
	int howmany = 4*Bout*Bout;
	int idist = 2*Bout, odist = 2*Bout;
	int rank = 2 ;
	int inembed[2] = {2*Bout,4*Bout*Bout};
	int onembed[2] = {2*Bout,4*Bout*Bout};
	int istride = 1, ostride = 1 ;
	int na[2] = {1,2*Bout};

	p1 = fftw_plan_many_dft( rank, na, howmany,
	                                   reinterpret_cast<fftw_complex*> (workspace1),
									   inembed, istride, idist,
									   reinterpret_cast<fftw_complex*> (&so3Sig.get(0)),
									   onembed, ostride, odist,
				                       FFTW_FORWARD, FFTW_ESTIMATE );

	seminaive_naive_table
	          =           SemiNaive_Naive_Pml_Table(B, B, seminaive_naive_tablespace, (double *) workspace2);
	trans_seminaive_naive_table
	          = Transpose_SemiNaive_Naive_Pml_Table(seminaive_naive_table, B, B, trans_seminaive_naive_tablespace, (double *) workspace2);

	// LOTS OF SPACE for precomputed wigners
	int wignerSpace = ((Bout*Bout)*(2+3*Bout+Bout*Bout))/3 ;
	wignersTrans = new double[wignerSpace];
  	genWigAllTrans( Bout, wignersTrans, workspace3 ) ;
}

void cleanup() {
	delete[] workspace2, workspace3;
	delete[] rdata, idata;
	delete[] so3Coef, seminaive_naive_tablespace, weights;
	delete[] trans_seminaive_naive_tablespace;
	delete[] wignersTrans;
}


//
// spherical harmonic transform using S2kit
//   assumes space already allocated
//
void SOFT_sharm_transform(map_3d<double> &sigR, map_3d<double> &sigCoefR, map_3d<double> &sigCoefI) {
	int R = sigR.getDims(0);

	for (int r_idx=0; r_idx<R; ++r_idx) {
		memcpy( rdata , &sigR.get(r_idx,0,0) , sizeof(double)*4*B*B );

		// idata should always == 0, but sometimes
		//    invtransform leaves it in a weird state
		for (int i=0; i<4*B*B; i++) idata[i] = 0;

		FST_semi_memo( rdata, idata,
					  &sigCoefR.get(r_idx,0,0),
					  &sigCoefI.get(r_idx,0,0),
					   B, seminaive_naive_table,
					   (double *) workspace2, 1, B ,
					  &dctPlan, &fftPlan, weights) ;
	}
}

//
// spherical harmonic transform using S2kit
//   assumes space already allocated
//
void SOFT_sharm_invTransform(map_3d<double> &sigR, map_3d<double> &sigCoefR, map_3d<double> &sigCoefI) {
	int R = sigR.getDims(0);

	for (int r_idx=0; r_idx<R; ++r_idx) {
		InvFST_semi_memo( &sigCoefR.get(r_idx,0,0), &sigCoefI.get(r_idx,0,0),
						rdata, idata,
						B, trans_seminaive_naive_table,
						(double *) workspace2, 0, B ,
						&idctPlan, &ifftPlan) ;
		memcpy(  &sigR.get(r_idx,0,0), rdata , sizeof(double)*4*B*B );
	}
}

void SOFT_Kazhdan_rep(map_3d<double> &sigR, vector< double > &sig_rotinv) {
	int R = sigR.getDims(0);
	sig_rotinv.resize(R*B);
	map_3d<double> sigCoefR(R,B,B), sigCoefI(R,B,B);

	SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);

	for(int rad_idx = 0; rad_idx < R; rad_idx++){
		for(int l = 0; l < B; l++){
			map_3d<double> tempR(1,B,B,0.0), tempI(1,B,B,0.0);
			for(int m = -l; m<= l; m++){
				int idx1 = rad_idx*B*B + seanindex(m,l,B);
				int idx2 = seanindex(m,l,B);
				tempR.set(idx2,sigCoefR.get(idx1));
				tempI.set(idx2,sigCoefI.get(idx1));
			}
			InvFST_semi_memo( &tempR.get(0,0,0), &tempI.get(0,0,0),
					rdata, idata,
					B, trans_seminaive_naive_table,
					(double *) workspace2, 0, B ,
					&idctPlan, &ifftPlan) ;

			map_3d<double>sigfl(1,sigR.getDims(1),sigR.getDims(2));
			memcpy(  &sigfl.get(0,0,0), rdata , sizeof(double)*4*B*B );
			double sum = 0;
			for(int i = 0; i < sigfl.getDimsProd();i++){
				sum += sigfl.get(i)*sigfl.get(i);
			}
			sig_rotinv[rad_idx*B+l] = sqrt(sum);
		}
	}
}
void SOFT_Frome_rep(map_3d<double> &sigR, vector< double > &sig_rotinv) {
	int R = sigR.getDims(0);
	sig_rotinv.resize(R*B*(B+2)/2);
	map_3d<double> sigCoefR(R,B,B), sigCoefI(R,B,B);

	SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);
	for(int rad_idx = 0; rad_idx < R; rad_idx++){
		for(int l = 0; l < B; l++){
			for(int m = 0; m<=l; m++){
				int idx = rad_idx*B*B + seanindex(m,l,B);
				sig_rotinv[rad_idx*B*(B+2)/2+l*(l+2)/2+m] = sqrt(pow(sigCoefR.get(idx),2) + pow(sigCoefI.get(idx),2)) ;
			}
		}
	}
}

//
// return a rotation-invariant representation of some region of density
//
void SOFT_rot_inv_rep(map_3d<double> &sigR, vector< double > &sig_rotinv) {
	int R = sigR.getDims(0);
	sig_rotinv.resize(R*B*B*B);

	map_3d<double> sigCoefR(R,B,B), sigCoefI(R,B,B);

	SOFT_sharm_transform(sigR, sigCoefR, sigCoefI);

	int idx1,idx2,idx3;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		//...
		for (int l=0; l<B; ++l)
		for (int l1=0; l1<B; ++l1)
		for (int l2=0; l2<B; ++l2) {

			complex<double> outersum(0,0);

			for (int m=-l; m<=l; ++m) {
				complex<double> innersum(0,0);

				for (int m1 = max(-l1,m-l2); m1 <= min(l1,m+l2) ; ++m1) {
					idx1 = r_idx*B*B + seanindex(m1,l1,B);
					idx2 = r_idx*B*B + seanindex(m-m1,l2,B);
					innersum += ClebschGordan(l1,l2,l,m1,m-m1,m)
					              * complex<double>( sigCoefR.get( idx1 ) , -sigCoefI.get( idx1 ) )
					              * complex<double>( sigCoefR.get( idx2 ) , -sigCoefI.get( idx2 ) );

				}

				idx3 = r_idx*B*B + seanindex(m,l,B);

				outersum += innersum * complex<double>( sigCoefR.get( idx3 ) , sigCoefI.get( idx3 ) );
			}

			sig_rotinv[ r_idx*B*B*B + l*B*B + l1*B + l2 ] = outersum.real();
		}
	}
}

//
// bandwidth-limit signal
//
void SOFT_bandwidth_limit(map_3d<double> &sigR) {
	int R = sigR.getDims(0);

	map_3d<double> sigCoefR(R,B,B), sigCoefI(R,B,B);

	SOFT_sharm_transform   (sigR , sigCoefR , sigCoefI);
	SOFT_sharm_invTransform(sigR , sigCoefR , sigCoefI);
}




//
//
//
void sph_standardize(map_3d<double> &sigR, map_3d<double> &maskR) {
	int R = sigR.getDims(0);
	double thisRWt;
	double wt=0, s=0, s2=0;
	double theta, thisWt;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		thisRWt =  ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );

		for (int th = 0; th < 2*B; ++th) {
			theta  = (2*th+1)*M_PI/(4*B);
			thisWt = thisRWt * sin(theta);

			for (int ph = 0; ph<2*B; ++ph) {
				wt += thisWt * maskR.get(r_idx,th,ph) ;
				s  += thisWt * maskR.get(r_idx,th,ph) * sigR.get(r_idx,th,ph);
				s2 += thisWt * maskR.get(r_idx,th,ph) * square( sigR.get(r_idx,th,ph) );
			}
		}
	}

	s = s/wt;
	s2 =  sqrt(s2/wt - s*s);

	sigR.scaleShift( 1/s2 , -s/s2 );
}

//
// standardize the signal given spharm coefficients
//
void sph_standardize_fromCoef(map_3d<double> &sigCoefR, map_3d<double> &sigCoefI) {
	int R = sigCoefR.getDims(0);
	int B = sigCoefR.getDims(1);

	double thisRWt, sumCoef2;
	double b_scaleFactor = sqrt(1.0/sqrt(0.015*B)); // this normalizes map s.t. correlation with itself == 1

	double s=0,s2=0, wt=0;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		// standardize
		thisRWt =  M_PI * ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );
		//sigCoefR.set(r_idx,0,0, 0.0); // set mean to 0

		sumCoef2 = 0.0;
		for (int i=0; i<B; ++i)
		for (int j=0; j<B; ++j)
			if (i != 0 || j!= 0)
				sumCoef2 += square(sigCoefR.get(r_idx,i,j)) + square(sigCoefI.get(r_idx,i,j));

		//thisScaleFactor = thisRWt * b_scaleFactor / sqrt ( sumCoef2 ) ;
		wt += thisRWt;
		s  += thisRWt * sigCoefR.get(r_idx,0,0);
		s2 += thisRWt * sumCoef2;
	}

	s = s/wt;
	s2 = sqrt(2 * b_scaleFactor * wt / s2);

	for (int r_idx=0; r_idx<R; ++r_idx) {
		for (int i=0; i<B; ++i)
		for (int j=0; j<B; ++j) {
			if (i == 0 && j== 0) {
				// adjust mean
				sigCoefR.set(r_idx,0,0 ,-s + sigCoefR.get(r_idx,0,0) );
				sigCoefI.set(r_idx,0,0 , 0.0 );
			} else {
				// adjust stdev
				sigCoefR.set(r_idx,i,j , s2 * sigCoefR.get(r_idx,i,j) );
				sigCoefI.set(r_idx,i,j , s2 * sigCoefI.get(r_idx,i,j) );
			}
		}
	}
}

//
//
//
void sph_standardize(map_3d<double> &sigR, int Rmax=-1) {
	int R = sigR.getDims(0);
	if (Rmax > 0) R = min(R,Rmax);

	double thisRWt;
	double wt=0, s=0, s2=0;
	double theta, thisWt;


	for (int r_idx=0; r_idx<R; ++r_idx) {
		thisRWt = ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );

		for (int th = 0; th < 2*B; ++th) {
			theta  = (2*th+1)*M_PI/(4*B);
			thisWt = thisRWt * sin(theta);

			for (int ph = 0; ph<2*B; ++ph) {
				wt += thisWt ;
				s  += thisWt * sigR.get(r_idx,th,ph);
				s2 += thisWt * square( sigR.get(r_idx,th,ph) );
			}
		}
	}

	s = s/wt;
	s2 =  sqrt(s2/wt - s*s);

	sigR.scaleShift( 1/s2 , -s/s2 );

}

//
// get the error by BW-limiting the signal
//
void SOFT_transform_error(map_3d<double> sigR, double &cumulative_error) {
	int R = sigR.getDims(0);
	cumulative_error = 0.0;

	map_3d<double> bwlimSigR(R,2*B,2*B);
	map_3d<double> sigCoefR(R,B,B, 0.0), sigCoefI(R,B,B, 0.0);

	map_3d<double> newSigR = sigR;
	sph_standardize(newSigR);

	SOFT_sharm_transform   (newSigR  , sigCoefR , sigCoefI);
	SOFT_sharm_invTransform(bwlimSigR , sigCoefR , sigCoefI);

	double wtSum = 0.0;

	for (int i=0; i< R; ++i) {
		double thisRWt =  ( 4.0 * square((i+1)*Rstep) + 1.0/3.0 );

		for (int th = 0; th < 2*B; ++th) {
			double theta  = (2*th+1)*M_PI/(4*B);
			double thisWt = thisRWt * sin(theta);

			for (int ph = 0; ph<2*B; ++ph) {
				cumulative_error += thisWt * square( newSigR.get(i,th,ph) - bwlimSigR.get(i,th,ph) );

				wtSum += thisWt;
			}
		}
	}

	cumulative_error /= wtSum;
}



//
//
//
void SOFT_so3_rotate( map_3d<double> &RsigCoefR, map_3d<double> &RsigCoefI ,
                      map_3d<double>  &sigCoefR, map_3d<double>  &sigCoefI ,
					  double alpha, double beta, double gamma) {
	int R = sigCoefR.getDims(0);
	for (int r_idx=0; r_idx<R; ++r_idx) {
		// clear ws 2 (is this needed?)
		for (int i=0; i<14*B*B + 48*B; ++i) workspace2[i] = complex<double>(0,0);
		rotateCoefAll( B, B, B-1,
		               alpha, beta, gamma,
		               &sigCoefR.get(r_idx,0,0),  &sigCoefI.get(r_idx,0,0),
		               &RsigCoefR.get(r_idx,0,0), &RsigCoefI.get(r_idx,0,0),
		               (double*) workspace2 );
	}
}



//
//
//
void SOFT_so3_correlate(
                 map_3d<double> &sigCoefR, map_3d<double> &sigCoefI,
				 map_3d<double> &tmpCoefR, map_3d<double> &tmpCoefI,
				 int Rmax = -1)
{
	int R = sigCoefR.getDims(0);

	if (Rmax > 0) R = min(R,Rmax);

	double thisRWt, sumRWt = 0.0;
	for (int r_idx=0; r_idx<R; ++r_idx)
		sumRWt += M_PI * ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );

	so3SigSum.setAll(complex<double>(0.0,0.0));
	for (int r_idx=0; r_idx<R; ++r_idx) {
		so3CombineCoef_fftw(
				B, Bout, Bout-1,
				&sigCoefR.get(r_idx,0,0), &sigCoefI.get(r_idx,0,0),
				&tmpCoefR.get(r_idx,0,0), &tmpCoefI.get(r_idx,0,0),
				reinterpret_cast<fftw_complex*>(so3Coef) ) ;

		// inverse so(3)
		Inverse_SO3_Naive_fftw_pc(
					Bout,
					reinterpret_cast<fftw_complex*> (so3Coef),
					reinterpret_cast<fftw_complex*> (&so3Sig.get(0)),
					reinterpret_cast<fftw_complex*> (workspace1),
					reinterpret_cast<fftw_complex*> (workspace2),
					workspace3, &p1, wignersTrans, 1 ) ;

		thisRWt = M_PI * ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 ) / sumRWt;
		so3Sig.scale(thisRWt);

		map_plus(so3SigSum, so3SigSum, so3Sig);
	}
}



//
//
//
void SOFT_so3_masked_correlate(
                  map_3d<double> &sigCoefR , map_3d<double> &sigCoefI ,
 				  map_3d<double> &tmpCoefR , map_3d<double> &tmpCoefI ,
                  map_3d<double> &sig2CoefR, map_3d<double> &sig2CoefI,
				  map_3d<double> &mskCoefR , map_3d<double> &mskCoefI , double minAllowedStdev)
{
	int R = sigCoefR.getDims(0);

	double sumMaskRWt=0.0, sumRWt=0.0;
	for (int r_idx=0; r_idx<R; ++r_idx) {
		sumRWt     +=                                            M_PI * ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );
		sumMaskRWt += (mskCoefR.get(r_idx,0,0) / sqrt(4*M_PI)) * M_PI * ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );
	}

	map_3d< double > meanSig(2*B,2*B,2*B), stdevSig(2*B,2*B,2*B), convolution(2*B,2*B,2*B);

	SOFT_so3_correlate(sigCoefR, sigCoefI , mskCoefR, mskCoefI);
	for (int i=0; i<8*B*B*B; ++i)
		meanSig.set(i, so3SigSum.get(i).real());

	SOFT_so3_correlate(sig2CoefR, sig2CoefI , mskCoefR, mskCoefI);
	double minComputedStdev = INFINITY, thisStdev, shiftStdev;
	for (int i=0; i<8*B*B*B; ++i) {
		thisStdev = 0.5 * sqrt(2.0) * so3SigSum.get(i).real() - 0.5 * square(meanSig.get(i));
		stdevSig.set(i, thisStdev);
		minComputedStdev = min(minComputedStdev , thisStdev);
	}

	// correct for errors at low BW
	minComputedStdev = min(minComputedStdev , minAllowedStdev);
	shiftStdev = minAllowedStdev - minComputedStdev;

	SOFT_so3_correlate(sigCoefR, sigCoefI , tmpCoefR, tmpCoefI);
	double thisCorrel, rWtScale = sumRWt/sumMaskRWt;

	for (int i=0; i<8*B*B*B; ++i) {
		thisCorrel = so3SigSum.get(i).real();
		so3SigSum.set(i, complex<double>(rWtScale * thisCorrel / ( sqrt( stdevSig.get(i) + shiftStdev )) , 0) );
	}
}



//
// correlate (masked) w/o rotating!
double SOFT_S2_correlate( map_3d<double> &sigCoefR, map_3d<double> &sigCoefI,
				        map_3d<double> &tmpCoefR, map_3d<double> &tmpCoefI,
				        map_3d<double> &mskCoefR, map_3d<double> &mskCoefI)
{
	int R = sigCoefR.getDims(0);
	map_3d<double> sigR(R,2*B,2*B), tmpR(R,2*B,2*B), mskR(R,2*B,2*B);

	SOFT_sharm_invTransform(sigR, sigCoefR, sigCoefI);
	SOFT_sharm_invTransform(tmpR, tmpCoefR, tmpCoefI);
	SOFT_sharm_invTransform(mskR, mskCoefR, mskCoefI);

	// RSCC
	double thisWt, thisRWt, theta;
	double wt=0.0, sig=0.0, tmp=0.0, sig2=0.0, tmp2=0.0, sigtmp=0.0;
	for (int r_idx=0; r_idx<R; ++r_idx) {
		thisRWt =  ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );

		for (int th = 0; th < 2*B; ++th) {
			theta  = (2*th+1)*M_PI/(4*B);
			thisWt = thisRWt * sin(theta);

			for (int ph = 0; ph<2*B; ++ph) {
				if (mskR.get(r_idx,th,ph) > 0.5) {
					wt += thisWt ;
					sig  += thisWt * sigR.get(r_idx,th,ph);
					sig2 += thisWt * square( sigR.get(r_idx,th,ph) );
					tmp  += thisWt * tmpR.get(r_idx,th,ph);
					tmp2 += thisWt * square( tmpR.get(r_idx,th,ph) );
					sigtmp += thisWt * sigR.get(r_idx,th,ph) * tmpR.get(r_idx,th,ph);
				}
			}
		}
	}

	sig /= wt;
	tmp /= wt;

	double ss_sigsig = sig2/wt - sig*sig;
	double ss_tmptmp = tmp2/wt - tmp*tmp;
	double ss_sigtmp = sigtmp/wt - sig*tmp;

	return ( ss_sigtmp / sqrt(ss_sigsig * ss_tmptmp) );
}



//
// correlate w/o rotating! v2.0 does not perform inverse transform
double SOFT_S2_correlate( map_3d<double> &sigCoefR, map_3d<double> &sigCoefI,
				          map_3d<double> &tmpCoefR, map_3d<double> &tmpCoefI)
{
	int R = sigCoefR.getDims(0);

	// RSCC
	double thisWt, wt=0.0, sig=0.0, tmp=0.0, sig2=0.0, tmp2=0.0, sigtmp=0.0;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		thisWt =  ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );
		wt += thisWt;

		sig += thisWt * ( square( sigCoefR.get(r_idx, 0,0) ) );
		tmp += thisWt * ( square( tmpCoefR.get(r_idx, 0,0) ) );

		for (int i = 0; i < B; ++i)
		for (int j = 0; j < B; ++j) {
			if (i==0 && j==0) continue;

			sig2 += thisWt * ( square( sigCoefR.get(r_idx, i,j) ) + square( sigCoefI.get(r_idx, i,j) ) );
			tmp2 += thisWt * ( square( tmpCoefR.get(r_idx, i,j) ) + square( tmpCoefI.get(r_idx, i,j) ) );
			sigtmp += thisWt * (   sigCoefR.get(r_idx, i,j) * tmpCoefR.get(r_idx, i,j)
			                     + sigCoefI.get(r_idx, i,j) * tmpCoefI.get(r_idx, i,j) ) ;
		}
	}

	sig /= wt;
	tmp /= wt;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		thisWt =  ( 4.0 * square((r_idx+1)*Rstep) + 1.0/3.0 );

		sig2 += thisWt * ( square( sigCoefR.get(r_idx, 0,0) - sig ) );
		tmp2 += thisWt * ( square( tmpCoefR.get(r_idx, 0,0) - tmp ) );
		sigtmp += thisWt * ( ( sigCoefR.get(r_idx, 0,0) - sig ) * ( tmpCoefR.get(r_idx, 0,0) - tmp ) ) ;
	}

	return ( sigtmp / sqrt(sig2 * tmp2) );
}



//
//
//
void SOFT_idx_to_rot(int maxloc , mat33<double> &thisRot) {
	int ii = maxloc / (4*Bout*Bout);
	int tmp = maxloc - (ii*4*Bout*Bout);
	int jj = tmp / (2*Bout);
	tmp = maxloc - (ii*4*Bout*Bout) - jj*(2*Bout);
	int kk = tmp ;

	double alpha = M_PI*kk/((double) Bout);
	double beta  = M_PI*(2*ii+1)/(4.0*Bout);
	double gamma = M_PI*jj/((double) Bout);
	thisRot = mat33<double>(alpha,beta,gamma);
}

void SOFT_idx_to_euler(int maxloc , vec3<double> &eulerAngles) {
	int ii = maxloc / (4*Bout*Bout);
	int tmp = maxloc - (ii*4*Bout*Bout);
	int jj = tmp / (2*Bout);
	tmp = maxloc - (ii*4*Bout*Bout) - jj*(2*Bout);
	int kk = tmp ;

	double alpha = M_PI*kk/((double) Bout);
	double beta  = M_PI*(2*ii+1)/(4.0*Bout);
	double gamma = M_PI*jj/((double) Bout);

	eulerAngles = vec3<double>(alpha,beta,gamma);
}


// make this map a potential function (used in BP)
void cache_spherical_resampling_linear(map_3d< linearCoeff3d > &ptrs , vec3<int> dims, Crystal &crystData) {
	int R = ptrs.getDims(0);

	vec3<double> cartX, fracX;
	linearCoeff3d idxXWts;
	vec3<int> idxX;
	double theta, phi, r;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		r = (double) Rstep*(r_idx+1.0);
		for (int th_idx=0; th_idx<2*B; ++th_idx) {
			theta = (2.0*th_idx + 1.0) * M_PI / (4.0*B);
			for (int phi_idx=0; phi_idx<2*B; ++phi_idx) {
				phi = (2.0*phi_idx) * M_PI / (2.0*B);

				cartX[0] = r*cos(phi)*sin(theta);
				cartX[1] = r*sin(phi)*sin(theta);
				cartX[2] = r*cos(theta);

				fracX = cartX.apply(crystData.c2f);

				// reverse X/Y??
				idxX[1] = (int) floor(fracX[0]*dims[0]);
				idxX[0] = (int) floor(fracX[1]*dims[1]);
				idxX[2] = (int) floor(fracX[2]*dims[2]);

				// wts
				double fx = fracX[1]*dims[1] - idxX[0],
				       fy = fracX[0]*dims[0] - idxX[1],
					   fz = fracX[2]*dims[2] - idxX[2];

				idxXWts.ptLower = idxX;
				idxXWts.wts[0] = (1.0-fx)*(1.0-fy)*(1.0-fz);
				idxXWts.wts[1] = (1.0-fx)*(1.0-fy)*     fz ;
				idxXWts.wts[2] = (1.0-fx)*     fy *(1.0-fz);
				idxXWts.wts[3] = (1.0-fx)*     fy *     fz ;
				idxXWts.wts[4] =      fx *(1.0-fy)*(1.0-fz);
				idxXWts.wts[5] =      fx *(1.0-fy)*     fz ;
				idxXWts.wts[6] =      fx *     fy *(1.0-fz);
				idxXWts.wts[7] =      fx *     fy *     fz;

				ptrs.set(r_idx, th_idx, phi_idx  ,  idxXWts);
			}
		}
	}
}

void cache_spherical_resampling_cubic_spline(map_3d< cubicSplineCoeff3d > &ptrs , vec3<int> dims, Crystal &crystData) {
	int R = ptrs.getDims(0);

	vec3<double> cartX, fracX, X;
	cubicSplineCoeff3d idxXWts;
	double theta, phi, r, w;
	int pt;

	for (int r_idx=0; r_idx<R; ++r_idx) {
		r = (double) Rstep*(r_idx+1.0);
		for (int th_idx=0; th_idx<2*B; ++th_idx) {
			theta = (2.0*th_idx + 1.0) * M_PI / (4.0*B);
			for (int phi_idx=0; phi_idx<2*B; ++phi_idx) {
				phi = (2.0*phi_idx) * M_PI / (2.0*B);

				// reverse X/Y??  This works so don't fuck with it
				cartX[1] = r*cos(phi)*sin(theta);
				cartX[0] = r*sin(phi)*sin(theta);
				cartX[2] = r*cos(theta);

				fracX = cartX.apply(crystData.c2f);

				X[0] = (fracX[0]*dims[0]);
				X[1] = (fracX[1]*dims[1]);
				X[2] = (fracX[2]*dims[2]);

				// wts
				for (int dim=0; dim<3; dim++) {
					// compute ULR interpolation index
					idxXWts.idx[dim] = (int)floor( X[dim] ) - 1;

					// compute the interpolation weights
					w = X[dim] - (double)(idxXWts.idx[dim]+1);
					idxXWts.wt[dim][3] = (1.0 / 6.0) * w * w * w;
					idxXWts.wt[dim][0] = (1.0 / 6.0) + (1.0 / 2.0) * w * (w - 1.0) - idxXWts.wt[dim][3];
					idxXWts.wt[dim][2] = w + idxXWts.wt[dim][0] - 2.0 * idxXWts.wt[dim][3];
					idxXWts.wt[dim][1] = 1.0 - idxXWts.wt[dim][0] - idxXWts.wt[dim][2] - idxXWts.wt[dim][3];
				}

				ptrs.set(r_idx, th_idx, phi_idx  ,  idxXWts);
			}
		}
	}
}


void spherical_resample_linear(
                 map_3d< double >     &sigR,
				 map_3d< double >     &fragMap,
				 map_3d< linearCoeff3d > &ptrs,
				 vec3<int>             centerAt,
				 int                   maxR = -1) {
	vec3<int> dims = fragMap.getDims();
	vec3<int> Y;
	double sum=0.0, sum2=0.0, N=0.0, wt, r, theta, thisVal;
	int R = ptrs.getDims(0),
	    n = ptrs.getDims(2);

	if (maxR > 0) R = min(R, maxR);

	for (int r_idx=0; r_idx<R; ++r_idx) {
		r = (double) Rstep*(r_idx+1.0);
		for (int th_idx=0; th_idx<n; ++th_idx) {
			theta = (2.0*th_idx + 1.0) * M_PI / (2.0*n);

			for (int phi_idx=0; phi_idx<n; ++phi_idx) {
				thisVal = 0;

				linearCoeff3d &idxX = ptrs.get(r_idx,th_idx,phi_idx);
				for (int i=0; i<2; ++i)
				for (int j=0; j<2; ++j)
				for (int k=0; k<2; ++k) {
					Y = centerAt + idxX.ptLower + vec3<int>(i,j,k); Y.mod(dims);
					thisVal += idxX.wts[4*i+2*j+k] * fragMap.get(Y[0],Y[1],Y[2]);
				}

				sigR.set(r_idx,th_idx,phi_idx , thisVal);
			}
		}
	}
}


//
// spherical resampling around a grid point; precompute coefficients
void spherical_resample_cubic_spline(
                 map_3d< double >             &sigR,
				 map_3d< double >             &fragMap,
				 map_3d< cubicSplineCoeff3d > &ptrs,
				 vec3<int>                     centerAt,
				 int                           maxR = -1) {
	vec3<int> dims = fragMap.getDims();
	vec3<int> Y;
	double sum=0.0, sum2=0.0, N=0.0, wt, r, theta, thisVal, sum_k, sum_jk;
	int R = ptrs.getDims(0),
	    n = ptrs.getDims(2);
	int i,j,k, x;
	int interpIndices[3][4];

	if (maxR > 0) R = min(R, maxR);

	fragMap.computeSplineCoeffs(3);   // recomputes "on demand"
	double* Bcoeff = fragMap.getSplineCoeffs();

	for (int r_idx=0; r_idx<R; ++r_idx) {
		for (int th_idx=0; th_idx<n; ++th_idx) {
			for (int phi_idx=0; phi_idx<n; ++phi_idx) {
				thisVal = 0;

				cubicSplineCoeff3d &idxX = ptrs.get(r_idx,th_idx,phi_idx);

				// enforce periodic boundary conditions
				for (int dim = 0; dim < 3; ++dim) {
					x = pos_mod( centerAt[dim] + idxX.idx[dim], dims[dim]);
					interpIndices[dim][0] = x; if (++x == dims[dim]) x=0;
					interpIndices[dim][1] = x; if (++x == dims[dim]) x=0;
					interpIndices[dim][2] = x; if (++x == dims[dim]) x=0;
					interpIndices[dim][3] = x;
				}


				for (i = 0; i <= 3; i++) {
					sum_jk = 0.0;
					for (j = 0; j <= 3; j++) {
						sum_k = 0.0;
						for (k = 0; k <= 3; k++) {
							sum_k += idxX.wt[2][k] * Bcoeff[  interpIndices[0][i]*dims[1]*dims[2]
							                                + interpIndices[1][j]*dims[2]
							                                + interpIndices[2][k]  ];
						}
						sum_jk += idxX.wt[1][j] * sum_k;
					}
					thisVal += idxX.wt[0][i] * sum_jk;
				}

				sigR.set(r_idx,th_idx,phi_idx , thisVal);
			}
		}
	}
}



//
// spherical resampling around an ARBITRARY point, no precomputation of coefficients
void spherical_resample_linear(
                 map_3d< double >  &sigSph,
				 map_3d< double >  &sigCart,
				 Crystal &crystData,
				 vec3<double>        centerIdx,
				 double             Rstep) {
	int R = sigSph.getDims(0);
	int B = sigSph.getDims(1)/2;

	vec3<double> cartX, fracX, X;
	double theta, phi, r, w;
	int pt;

	vec3<double> centerFrac( centerIdx[0]/sigCart.getDims(0) ,
	                         centerIdx[1]/sigCart.getDims(1) ,
	                         centerIdx[2]/sigCart.getDims(2) );

	for (int r_idx=0; r_idx<R; ++r_idx) {
		r = (double) Rstep*(r_idx+1.0);
		for (int th_idx=0; th_idx<2*B; ++th_idx) {
			theta = (2.0*th_idx + 1.0) * M_PI / (4.0*B);
			for (int phi_idx=0; phi_idx<2*B; ++phi_idx) {
				phi = (2.0*phi_idx) * M_PI / (2.0*B);

				cartX[1] = r*cos(phi)*sin(theta);
				cartX[0] = r*sin(phi)*sin(theta);
				cartX[2] = r*cos(theta);

				fracX = cartX.apply(crystData.c2f);

				X[0] = pos_mod(fracX[0] + centerFrac[0], 1.0);
				X[1] = pos_mod(fracX[1] + centerFrac[1], 1.0);
				X[2] = pos_mod(fracX[2] + centerFrac[2], 1.0);

				//cout << "     ... interp " << X.to_string() << "!\n";
				sigSph.set(r_idx, th_idx, phi_idx  ,  sigCart.interp_linear(X));
			}
		}
	}
}


//
// spherical resampling around an ARBITRARY point, no precomputation of coefficients
void spherical_resample_cubic_spline(
                 map_3d< double > &sigSph,
				 map_3d< double > &sigCart,
				 vec3<double>       centerFrac,
				 Crystal &crystData) {
	int R = sigSph.getDims(0);

	vec3<double> cartX, fracX, X;
	double theta, phi, r, w;
	int pt;

	sigCart.computeSplineCoeffs(3);   // recomputes "on demand"

	for (int r_idx=0; r_idx<R; ++r_idx) {
		r = (double) Rstep*(r_idx+1.0);
		for (int th_idx=0; th_idx<2*B; ++th_idx) {
			theta = (2.0*th_idx + 1.0) * M_PI / (4.0*B);
			for (int phi_idx=0; phi_idx<2*B; ++phi_idx) {
				phi = (2.0*phi_idx) * M_PI / (2.0*B);

				cartX[1] = r*cos(phi)*sin(theta);
				cartX[0] = r*sin(phi)*sin(theta);
				cartX[2] = r*cos(theta);

				fracX = cartX.apply(crystData.c2f);

				X[0] = pos_mod(fracX[0] + centerFrac[0], 1.0);
				X[1] = pos_mod(fracX[1] + centerFrac[1], 1.0);
				X[2] = pos_mod(fracX[2] + centerFrac[2], 1.0);

				//cout << "     ... interp " << X.to_string() << "!\n";
				sigSph.set(r_idx, th_idx, phi_idx  ,  sigCart.interp_spline(X));
			}
		}
	}
}
