// //////////////////////////////////////////////////////////////////////////
//
//     BP.cpp
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////

#include <time.h>    // benchmarking

#include <sys/stat.h>  // mkdir
#include <sys/types.h>
#include <dirent.h>

#include "BP.h"
#include "getopts.h"   // parse command line args
#include <queue>
//Handle priority function for guiding message passing
typedef pair<int, double> nodePQ;
class Prioritize {
public:
     int operator() ( const nodePQ & p1,
                      const nodePQ & p2 ) {
         return p1.second < p2.second;
     }
};
int char2digit(char ch){
	switch(ch){
		case '9': return 9;
		case '8': return 8;
		case '7': return 7;
		case '6': return 6;
		case '5': return 5;
		case '4': return 4;
		case '3': return 3;
		case '2': return 2;
		case '1': return 1;
		case '0': return 0;
	}
	return -1;
}
//
// main()
//
int main(int argc, char** argv) {
	version();
	///
	/// read input files
	Options bpOpts;
  	int switchInt;

	//                  Short   Long        Description           Take args?
	bpOpts.addOption("w",    "working",      "[REQUIRED] Working Directory",      true);
	bpOpts.addOption("c",    "cryst",        "[REQUIRED] Input crystal file",      true);
	bpOpts.addOption("s",    "seq",          "[REQUIRED] Input sequence file",      true);
	bpOpts.addOption("o",    "sigma",        "Average squared rotation error (default 1.0)",      true);
	bpOpts.addOption("m",    "m-est",        "Probability of incorrect rotational alignment (default 0.1)",      true);
	bpOpts.addOption("i",    "iters",        "Iterations of BP (default 40)",      true);
	bpOpts.addOption("e",    "prioratoms",    "PDB file containing locations of selenium and/or CA atoms",      true);
	bpOpts.addOption("p",    "se-prior",     "Weight for provided prior atom list (E.g. -e SeWgt CaWgt) default 0.9 0.1",      true);
	bpOpts.addOption("x",    "memory",       "Maximum memory usage before paging (in MB)",   true);
	bpOpts.addOption("d",    "scratch",      "Scratch directory for paging (will use a unique subdir)",      true);
	bpOpts.addOption("P",    "partial",      "Partial solution file",      true);
	bpOpts.addOption("F",    "fixed",        "Fix residues in partial solution",      false);
	bpOpts.addOption("s",    "secondary",    "Input secondary structure sequence file",      true);
	bpOpts.addOption("g",    "guide",    	 "Input priority function for guiding message passing",      true);
	bpOpts.addOption("S",    "start",		 "Central residue for beginning BP", true );
	bpOpts.addOption("C",    "continue",	 "Start from previous BP results (i.e., use previous posteriors as starting point",false );
	bpOpts.addOption("f",    "force",		 "Ignore previous checkpoint", false );
	bpOpts.addOption("B",    "backwards",	 "Process equally weighted nodes from right to left", false );
	bpOpts.addOption("t",    "checkpoint",   "Toggle checkpointing off", false );

	string workingdir, crystfile, seqfile, prioratomfile, ssfile, guidefile;
	double seleniumprior = 0.9, caprior = 0.1;
	int seed = 0;
	SCRATCH_DIR = DEFAULT_SCRATCH_DIR;
	MEM_LIMIT   = DEFAULT_MEM_LIMIT;
	SIGMA        = 1.0;
	M_EST_ORIENT = 0.1;
	MAX_ITERS    = 40;
	CAtrace priorStruct;
	bool doCheckPt = true, restart = false, force = false, fixBelief = false; //Whether to fix location of residues in the given partial solution
	int dir = 1;
	bpOpts.parse(argc, argv);
	while ((switchInt = bpOpts.cycle()) >= 0) {
		vector<string> pargs;
		Crystal nullCryst;
		switch(switchInt) {
		case 0:   // Option 'working-dir'
			workingdir =  bpOpts.getArgs(switchInt); break;
		case 1:   // Option 'cryst'
			crystfile =  bpOpts.getArgs(switchInt); break;
		case 2:   // Option 'seq'
			seqfile =  bpOpts.getArgs(switchInt); break;
		case 3:   // Option 'sigma'
			SIGMA = atof( bpOpts.getArgs(switchInt) ); break;
		case 4:   // Option 'm-est'
			M_EST_ORIENT = atof( bpOpts.getArgs(switchInt) ); break;
		case 5:   // Option 'iters'
			MAX_ITERS = atoi( bpOpts.getArgs(switchInt) ); break;
		case 6:  // Option 'prioratoms'
			prioratomfile =  bpOpts.getArgs(switchInt); break;
		case 7:  // Option 'se-prior'
			Tokenize( bpOpts.getArgs(switchInt) , pargs );
			if (pargs.size() == 2) {
				seleniumprior =  atof( pargs[0]);
				caprior = atof( pargs[1]);
			} else {
				cerr << "Error with prior atom list weights: two weights must be provided with Selenium prior first and CA prior second" << endl;
				exit(1);
			}
			break;
		case 8:   // Option 'memory'
			MEM_LIMIT = atof( bpOpts.getArgs(switchInt) ); break;
		case 9:   // Option 'scratchdir'
			SCRATCH_DIR = bpOpts.getArgs(switchInt); break;
		case 10:  // Option 'partial'
			priorStruct =  CAtrace(bpOpts.getArgs(switchInt),nullCryst,true);
			cerr << "Using Partial Solution.  WARNING: residues must be numbered according to sequence position.\n";
			cerr << "All pdb atoms should be part of chain A, resID is continuous (e.g. first res of chain B is resID = |A|+1).\n";
			cerr << "To lock in these residues, add -f flag.\n";
			break;
		case 11:  //Option 'fixed'
			fixBelief =  true; break;
		case 12:  //Option 'secondary'
			ssfile = bpOpts.getArgs(switchInt); break;
		case 13:  //Option 'secondary'
			guidefile = bpOpts.getArgs(switchInt); break;
		case 14:
			seed = atoi( bpOpts.getArgs(switchInt) ); break;
		case 15:
			restart = true;
			cerr << "Starting from previous run of BP" << endl;
			break;
		case 16:
			force = true;
			break;
		case 17:
			dir = -1;
			break;
		case 18:
			doCheckPt = false;
			break;
		}
	}

	// at a minimum, need to specify workingdir, map, cryst, seq
	if (workingdir.length() == 0 || crystfile.length() == 0 || seqfile.length() == 0) {
		cerr << "Error: Must specify a working directory, crystal file, and sequence file.\n";
		cerr << "       See \'" << argv[0] << " -h\' for more details\n";
		exit(1);
	}

	////////////////////
	//
	// read input files
	//
	Crystal crystData(crystfile);
	Sequence  seqData(seqfile);
	int nResSum = seqData.nres();
	vec3<int> dims = vec3<int>(0,0,0);

	string filestem = seqfile;
		filestem = filestem.substr(filestem.find_last_of('/')+1);
		filestem = filestem.substr(0,filestem.find_last_of('.'));
	string outpdbdir = workingdir + "/backbone_models/";
	string outmapdir = workingdir + "/posteriors/";
	string matchFormatString = workingdir + "/priors/";
	string recoverdir = outmapdir + "checkpoint/";

	mkdir(outpdbdir.c_str(), 0755);
	mkdir(outmapdir.c_str(), 0755);

	//////////////////////
	///
	///  pretty-print
	cerr << "Running with inputs: " << endl;
	cerr << "    working directory = \'" << workingdir << "\'" << endl;
	cerr << "    crystal dims      = " << crystData.to_string() << endl;
	cerr << "            spacegp   = " << crystData.symmGpId << endl;
	cerr << "    protein sequence  = " << seqData.to_string() << "(" << seqData.nres()<< " residues)" << endl;
	if (prioratomfile.length() > 0) {
		cerr <<  "    Prior Atom locations in   = " << prioratomfile << endl;
		cerr <<  "    SE prior prob.    = " << seleniumprior << endl;
		cerr <<  "    CA prior prob.    = " << caprior << endl;
	}

	cerr << endl;
	cerr << "Search parameters " << endl;
	cerr << "    sigma             = " << SIGMA << endl;
	cerr << "    M-estimate        = " << M_EST_ORIENT << endl;
	cerr << "    niters            = " << MAX_ITERS << endl;

	cerr << endl;

	////////////////////
	///
	///  if Se locations are provided, load them
	PriorAtomList PAs;
	if (prioratomfile.length() > 0) {
		PAs.load(prioratomfile);
		cerr << "Read " << PAs.nSe() << " seleniums and " << PAs.nCA() << " Calphas\n";
	}

	///////////////////
	///
	/// if secondary structure predictions are provided, load them
	Sequence sspred, guidefxn;
	if(ssfile.length() > 0){
		sspred.load(ssfile);
		if(sspred.nres() != nResSum){
			cerr << "Error: Secondary Structure predictions do not correspond to sequence length" << endl;
			exit(1);
		}
		for(int i = 0; i < nResSum; i++){
			int ss = ss_to_index(sspred.serialGet(i));
			if(ss >= NUM_SS || ss < 0){
				cerr << "Error: Invalid Secondary Structure Element at position " << i+1 << endl;
				exit(1);
			}
		}
	} else {
		for(int i = 0; i < seqData.nres(); i++){
			sspred.set(0,i,' ');
		}
	}


	if(guidefile.length() > 0){
			guidefxn.load(guidefile);
			if(guidefxn.nres() != nResSum){
				cerr << "Error: Guiding function do not correspond to sequence length" << endl;
				exit(1);
			}
	} else {
		for(int i = 0; i < seqData.nres(); i++){
			guidefxn.set(0,i,'9');
		}
	}
	////////////////////////////////////
	///
	/// Read match files
	///
	map< string, Belief > Prob0; // prior probabilities
	string mer;

	//
	// first get dims
	for (int i=0; i<nResSum; i++) {
		if (seqData.serialGet(i) == '_') continue;  // gap

		if(priorStruct.nres() > i && priorStruct.serialGet(i).typeShort != '_'){
 			mer = mer+itoc(i);
 		}
		mer = seqData.serialGet5mer(i);
		dims = Belief::getDims(matchFormatString, mer);
		break;
	}
	cerr << "    BP dims           = " << dims.to_string() << endl;
	crystData.buildMaps(dims);


	///////////////////////////////////
	///
	/// check if we need to page maps to disk
	///    --> if we DO, page that first map to disk
	double approxMemUsage = ( (nResSum*36.0) * crystData.getPtrMapInv().size() ) / (1024.0*1024.0) + 100.0;
	cerr << "Approximate memory needed: " << approxMemUsage << "MB\n";
	if (approxMemUsage > MEM_LIMIT) {
		PAGE_MAPS = true;
		bool dirCreated = false;
		// write to a unique sub-dir in SCRATCH_DIR
		srand(time(NULL));
		mkdir(SCRATCH_DIR.c_str(), 0755);   // ensure dirs exists
		string uniq = "";
		while(!dirCreated){
			uniq=".c"+itoc(rand());

			DIR *pdir;
			string dirName = SCRATCH_DIR+"/"+uniq+"/";
			pdir=opendir(dirName.c_str()); //"." refers to the current dir
			if (!pdir){
				dirCreated = true;
			}
			closedir(pdir);
		}
		SCRATCH_DIR = SCRATCH_DIR + "/" + uniq + "/";
		mkdir(SCRATCH_DIR.c_str(), 0755);

		cerr << "    page maps to \'" << SCRATCH_DIR << "\'\n";
	} else {
		cerr << "   keep maps in memory\n";
		PAGE_MAPS = false;
	}
	srand(1);

	////////////////////
	///
	/// cache LOCAL clique potentials psi_i->(i+1)
	///      and pairwise distances over the surface of a 3.8A sphere
	///
	getAdjPotentialList(dims,crystData);

	////////////////////
	///
	/// build cost matrix for each orientation-neighbor pair
	///      if no angular info is available, assign uniform probability across rotations
	///

	for(int ss=0; ss < NUM_SS; ss++){
		int n_adj_phis = phi_adj_idxs[ss].size(); //no. of possible next residue locations

		vector<float> uniform_cost(n_adj_phis * crystData.nSymm()),
					  uniform_inv_cost(n_adj_phis * crystData.nSymm());
		for (int j=0; j<n_adj_phis; j++) {
			uniform_cost[j] =    (1-M_EST_ORIENT) * phi_adj_phis[ss][j] + (M_EST_ORIENT/n_adj_phis);
			uniform_inv_cost[j] = log(   uniform_cost[j]   );

			for (int symm_idx=1; symm_idx < crystData.nSymm(); ++symm_idx) {
				uniform_cost[ symm_idx*n_adj_phis + j]     = uniform_cost[j];
				uniform_inv_cost[ symm_idx*n_adj_phis + j] = uniform_inv_cost[j];
			}
		}
		phi_cost_cache[ss][ 0 ] = uniform_cost;
		phi_inv_cost_cache[ss][ 0 ] = uniform_inv_cost;


		// otherwise, build an orientation specific cost function
		//    precompute for each possible orientation (on this grid)
		for (int i=0; i<n_adj_phis; i++) {
			vector<float> cost_xi(n_adj_phis * crystData.nSymm()),
						  inv_cost_xi(n_adj_phis * crystData.nSymm());

			double r_i = phi_adj_xs[ss][i].r(), r_j; //CA-CA length
			vec3<double> x_i, x_j;
			float dist_ij,cost_ij,dot_prod;

			x_i = phi_adj_xs[ss][i]/r_i;

			for (int symm_idx=0; symm_idx < crystData.nSymm(); ++symm_idx) {

				float cost_sum=0.0, inv_cost_sum=0.0;

				for (int j=0; j<n_adj_phis; j++) {
					r_j = phi_adj_xs[ss][j].r();
					x_j = (phi_adj_xs[ss][j]/r_j).apply( crystData.getSymmInvRot(symm_idx) );

					dot_prod = dot( x_i , x_j );
					if (dot_prod > 1)  dot_prod =  1;
					if (dot_prod < -1) dot_prod = -1;

					// forward
					dist_ij = CA_CA_DIST[ss] * acos( dot_prod );  // distance on sphere surface
					cost_ij = phi_adj_phis[ss][j] * exp( -square(dist_ij) / (2*square(SIGMA)) );
					cost_xi[ symm_idx*n_adj_phis + j ] = cost_ij ;
					cost_sum += cost_ij;

					// inverse
					dist_ij = CA_CA_DIST[ss] * acos( -dot_prod );  // distance on sphere surface
					cost_ij = phi_adj_phis[ss][j] * exp( -square(dist_ij) / (2*square(SIGMA)) );
					inv_cost_xi[ symm_idx*n_adj_phis + j ] = cost_ij ;
					inv_cost_sum += cost_ij;
				}

				// normalize, log-trasform inv_*
				for (int j=0; j<n_adj_phis; j++) {
					cost_xi[ symm_idx*n_adj_phis + j]     =       (1-M_EST_ORIENT) *     cost_xi[symm_idx*n_adj_phis +j]/    cost_sum
																+    M_EST_ORIENT  * uniform_cost[j];
					inv_cost_xi[ symm_idx*n_adj_phis + j] = log ( (1-M_EST_ORIENT) * inv_cost_xi[symm_idx*n_adj_phis +j]/inv_cost_sum
																+    M_EST_ORIENT  * uniform_cost[j] );

				}
			}

			int idx = to_index(phi_adj_idxs[ss][i] , dims);
			phi_cost_cache[ss][idx]     = cost_xi;
			phi_inv_cost_cache[ss][idx] = inv_cost_xi;
		}
	}
	vector<bool> priorSoln(nResSum);
	vector<bool> fixRes(nResSum);
	///
	/// Finally, read marginals
	///
	for (int i=0; i<nResSum; i++) {
		if (seqData.serialGet(i) == '_') continue;  // gap
		mer = seqData.serialGet5mer(i);

		// If we haven't seen this 5mer before, load it from disk
		if (Prob0.count(mer) == 0) {
			if(priorStruct.nres() > i && priorStruct.serialGet(i).typeShort != '_'){
				mer = mer+itoc(i);
				priorSoln[i] = true;
				if(fixBelief) fixRes[i] = true;
			} else {
				priorSoln[i] = fixRes[i] = false;
			}
			Prob0[mer] = Belief(matchFormatString, PAs, seleniumprior, caprior, mer, crystData, ss_to_index(sspred.serialGet(i)),priorSoln[i]);
			Prob0[mer].M.log_transform();

			if (PAGE_MAPS) Prob0[mer].dumpToDisk(SCRATCH_DIR+"/p0_"+mer);
		}
	}
	//Don't lock end points completely
	if(fixBelief){
		vector<bool> temp(nResSum);
		for(int i = 1; i < nResSum-1; i++){
			if(fixRes[i] && (!fixRes[i-1] || !fixRes[i+1])){
				temp[i] = false;
			} else {
				temp[i] = fixRes[i];
			}
		}
		fixRes = temp;
		fixRes[0] = fixRes[1];
		fixRes[nResSum-1] = fixRes[nResSum-2];
	}
	////////////////////
	//
	// Initialize model
	//      allocate space for belief, messages
	//      set initial belief to prior
	//      precompute backward & global msgs
	//
	cerr << "Initializing: \n";


	vector< compact_map_3d<float> >  b(nResSum);       // belief (projected)

	vector< AdjMessage >     msg_f(nResSum);   // forward  messages (from res i->(i+1))
	vector< AdjMessage >     msg_b(nResSum);   // backward messages (         i->(i-1))
	vector< compact_map_3d<float> >  msg_glob(nResSum); // global messages FROM res i(collision avoidance)

	compact_map_3d<double> msg_glob_prod;     // product of global messages
	msg_glob_prod.alloc(crystData, dims);
	msg_glob_prod.setAll(0.0);

	// temporary map variables
	compact_map_3d< float > b_i, b_j;
	int resId;
	char resType, chnId;
	bool recover = !force && exists(string(recoverdir+filestem+"_misc.txt"));
	for(int i = 0; i < nResSum && recover; i++){
		seqData.serialGet(i, resId, resType, chnId);
		recover = exists(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal.dat"));
	}

	// for each residue in each chain
	for (int i=0; i<nResSum; i++) {
		mer = seqData.serialGet5mer(i);
		cerr << mer[2]; cerr.flush();
		if (mer[2] == '_') continue;  // gap
		int ss = ss_to_index(sspred.serialGet(i));
		Belief *thisProb;
		if(priorSoln[i]){
			mer=mer+itoc(i);
		}
		thisProb = &Prob0[mer];

		//Look for checkpoints
		if (PAGE_MAPS) thisProb->loadFromDisk(SCRATCH_DIR+"/p0_"+mer);

		b[i].alloc(crystData,dims);
		if (restart || recover) {
			seqData.serialGet(i, resId, resType, chnId);
			string recovfile (filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal");
			recovfile = restart ? outmapdir + recovfile + ".mat" : recoverdir + recovfile + ".dat";
			if(restart){
				map_3d<float> post;
				post.readFromFile(recovfile.c_str());
				for (int j=0; j<post.getDimsProd(); ++j) {
					if (crystData.getAsymmUnit().get(j) > 0) {
						b[i].i_set(j, post.get(j));
					}
				}
			} else {
				b[i].dumpNoWrite();
				b[i].loadDataFromDisk(recovfile);
			}
			b[i].invlog_transform();
		} else {
			// set belief to prior
			for (int x=0; x<b[i].getNPts(); ++x) {
				b[i].d_set( x , (double) exp( thisProb->M.d_get(x) ) );
			}
		}

		// pre-compute backwards messages

		if (mer[1] != '_' && !(fixRes[i-1])) {
			if(recover){
				msg_b[i].M.alloc(crystData,dims);
				msg_b[i].phi.alloc(crystData,dims);
				msg_b[i].loadCheckPoint(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_b"));
			} else {
				compute_adj_message(msg_b[i], b[i], thisProb->phi_b, crystData, ss);
				msg_b[i].M.normalize( M_EST_POSITIONS ) ;
				msg_b[i].M.log_transform();
			}

			//msg_b[i].M.testIfContainsNaN("m_b_0");

			if (PAGE_MAPS) msg_b[i].dumpToDisk(SCRATCH_DIR+"/msgb"+itoc(i));
		}
		// pre-compute forwards messages

		if (mer[3] != '_' && !(fixRes[i+1])) {
			if(recover){
				msg_f[i].M.alloc(crystData,dims);
				msg_f[i].phi.alloc(crystData,dims);
				msg_f[i].loadCheckPoint(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_f"));
			} else {
				compute_adj_message(msg_f[i], b[i], thisProb->phi_f, crystData, ss);
				msg_f[i].M.normalize( M_EST_POSITIONS ) ;
				msg_f[i].M.log_transform();
			}
			//msg_b[i].M.testIfContainsNaN("m_b_0");

			if (PAGE_MAPS) msg_f[i].dumpToDisk(SCRATCH_DIR+"/msgf"+itoc(i));
		}
		if (PAGE_MAPS) thisProb->dumpNoWrite();

		// allocate space for global messages
		msg_glob[i].alloc(crystData , dims);
		if(recover){
			msg_glob[i].dumpNoWrite();
			msg_glob[i].loadDataFromDisk(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_glob.dat"));
		} else {
			msg_glob[i].setAll(0);
		}
		// log transform belief
		b[i].log_transform();
		//b[i].testIfContainsNaN("b_0");


		if (PAGE_MAPS) {
			msg_glob[i].dumpDataToDisk(SCRATCH_DIR+"/msgglob"+itoc(i)+".dat");
			b[i].dumpDataToDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
		}
	}

	// early trace??
	doTrace(string(outpdbdir+filestem+"_iter0.pdb"), b, seqData, crystData,sspred);

	//////////////////////////
	//
	// main loop!
	//

	//Create priority queue for ordering messages
	priority_queue< nodePQ, vector< nodePQ >, Prioritize> pq;
	vector<double> linear_pq(nResSum);
	vector<int> count(nResSum);



	double delta = INFINITY, RMSd;
	int iter=0, j_start, j_stop, j_inc;
	cerr << endl;
	vector<bool>updatef(nResSum), updateb(nResSum);
	for(int i=0;i<nResSum;i++){
		updatef[i]=updateb[i]=false;
	}
	if(recover) {
		//load global aggregator
		msg_glob_prod.dumpNoWrite();
		msg_glob_prod.loadDataFromDisk(string(recoverdir+filestem+"_msg_glob_prod.dat"));

		//load other state information
		fstream miscfile((string(recoverdir+filestem+"_misc.txt")).c_str());
		if(miscfile.is_open()){
			char str[10];
			miscfile >> str;
			iter = atoi(str);
			for(int i = 0; i < nResSum; i++){
				if(miscfile.bad() || miscfile.eof() || miscfile.fail()){
					cerr<<"Error reading misc. file at res " << i << endl;
					break;
				}
				//updateb,updatef,count
				char upb[10], upf[10];
				miscfile >> upb;
				updateb[i] = (atoi(upb) == 0)? false : true;
				miscfile >> upf;
				updatef[i] = (atoi(upf) == 0)? false : true;
				miscfile >> count[i] >> linear_pq[i];
			}
		} else {
			cerr << "Could not open misc. information file for recovery" << endl;
		}
		miscfile.close();
		cerr << "Recovering from iteration " << 1+iter/nResSum << endl;
	}
	for(int i=0; i<nResSum; i++){
		double nudge = 0;
		if(dir == 1)
			nudge = (double)(nResSum-((i-seed+nResSum)%nResSum))/nResSum/1000.0 ;
		else
			nudge = (double)i/nResSum/1000.0 ;
		if (!recover) linear_pq[i] = char2digit(guidefxn.serialGet(i))+nudge;
		pq.push(nodePQ(i,linear_pq[i]));
	}
	//Checkpoint before start
	mkdir(recoverdir.c_str(), 0755);
	if(!recover && doCheckPt){
		for(int j = 0; j < nResSum; j++){
			mer = seqData.serialGet5mer(j);
			if (b[j].getNPts() > 0) {
				seqData.serialGet(j, resId, resType, chnId);
				if (PAGE_MAPS){
					b[j].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
					msg_glob[j].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");
					if(mer[1] != '_' && !fixRes[j-1]) msg_b[j].loadFromDisk(SCRATCH_DIR+"/msgb"+itoc(j));
					if(mer[3] != '_' && !fixRes[j+1]) msg_f[j].loadFromDisk(SCRATCH_DIR+"/msgf"+itoc(j));
				}
				b[j].writeDataToDisk(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal.dat"));
				msg_glob[j].writeDataToDisk(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_glob.dat"));
				if(mer[1] != '_' && !fixRes[j-1]) msg_b[j].writeCheckPoint(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_b"));
				if(mer[3] != '_' && !fixRes[j+1]) msg_f[j].writeCheckPoint(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_f"));
				if (PAGE_MAPS){
					b[j].dumpNoWrite();
					msg_glob[j].dumpNoWrite();
					if(mer[1] != '_' && !fixRes[j-1]) msg_b[j].dumpNoWrite();
					if(mer[3] != '_' && !fixRes[j+1]) msg_f[j].dumpNoWrite();
				}
			}
			msg_glob_prod.writeDataToDisk(string(recoverdir+filestem+"_msg_glob_prod.dat"));
			fstream miscfile((string(recoverdir+filestem+"_misc.txt")).c_str(),ios::out);
			miscfile << iter << endl;
			for(int i = 0; i < nResSum; i++){
				//updateb,updatef,count
				miscfile << updateb[i] << " " << updatef[i] << " " << count[i] << " " << fixed << setprecision(7) << linear_pq[i] << endl;
			}
			miscfile.close();
		}
	}
	while (++iter <= MAX_ITERS*nResSum && (delta > 1e-4*nResSum || iter % nResSum != 1)) {   // limit # iterations
		if(recover && delta == INFINITY) delta = 0;

		if(iter%nResSum==1){
			if(delta != INFINITY) {
				cout << endl;
				cout << "Sum change in belief = " << delta << endl << endl;
				doTrace (string(outpdbdir+filestem+"_iter"+itoc(iter/nResSum)+".pdb"), b, seqData, crystData,sspred);
			}
			cerr << "Iteration " << 1+iter/nResSum << ":\n";
			delta = 0;
		}

		int j = pq.top().first;
		count[j]++;
		double tempPriority = pq.top().second;
		pq.pop();
		double nudge = 1.0/1000 - (tempPriority - floor(tempPriority));
		if(nudge<0) nudge = 0;
		if(nudge>1.0/1000) nudge = .001;
		linear_pq[j] = floor(tempPriority)-1+nudge;
		pq.push(nodePQ(j,linear_pq[j]));

		//
		// for j=1..nres, compute b[j] as product (actually sum-of-logs) of:
		//       M[j]        --  prior belief
		//       msg_f[j-1]  --  msg from neighbor j-1
		//       msg_b[j+1]  --  msg from neighbor j+1
		//       msg_glob_prod/(msg_glob[j-1]*msg_glob[j]*msg_glob[j+1])
		//                   --  combined msg from all non-adjacent neighbors
		//
		mer = seqData.serialGet5mer(j);
		cerr << mer[2]; cerr.flush();
		if (mer[2] == '_') continue;  // gap

		int ss = ss_to_index(sspred.serialGet(j));
		if(priorSoln[j]){
			mer=mer+itoc(j);
		}
		Belief& prob_j = Prob0[ mer ];
		if (PAGE_MAPS) prob_j.loadFromDisk(SCRATCH_DIR+"/p0_"+mer);
		if(fixRes[j])
		//Update belief
		{
			//load b(j) if fixed, no updating needed

			if (PAGE_MAPS){
				b[j].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
				msg_glob[j].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");
			}
			b_j = b[j];
			if (PAGE_MAPS) b[j].dumpNoWrite();
		} else {
			// set b(j) to global msg * prior
			// update global msg first to remove effect of res j
			if (PAGE_MAPS) msg_glob[j].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");
			map_minus(msg_glob_prod, msg_glob_prod, msg_glob[j]);
			map_plus( b_j, prob_j.M, msg_glob_prod );
			updatef[j]=updateb[j]=true;
		}
		if (mer[1] != '_' && !fixRes[j] && !(fixRes[j-1] && count[j] > 1) && updatef[j-1]) {
			int i=j-1;

			string mer_i = seqData.serialGet5mer(j-1);

			if (PAGE_MAPS) b[i].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
			b_i = b[i];
			if (PAGE_MAPS) b[i].dumpNoWrite();
			int ss_i = ss_to_index(sspred.serialGet(i));
			if(priorSoln[i]){
				mer_i=mer_i+itoc(i);
			}
			Belief* prob_i = &Prob0[ mer_i ];
			if (PAGE_MAPS && mer_i != mer) {
				prob_i->loadFromDisk(SCRATCH_DIR+"/p0_"+mer_i);
			}


			// page in
			if (PAGE_MAPS) {
				msg_b[j].loadFromDisk(SCRATCH_DIR+"/msgb"+itoc(j));
				msg_f[i].realloc();   // don't care what it is, just reallocate space
			}

			if(!fixRes[i]){
				belief_subtract_and_proj(b_i, prob_i->phi_f, msg_b[j], ss_i);
			}
			b_i.invlog_transform();

			compute_adj_message(msg_f[i], b_i, prob_i->phi_f, crystData, ss_i);
			msg_f[i].M.normalize(  M_EST_POSITIONS  );
			msg_f[i].M.log_transform();

			// page out
			if (PAGE_MAPS) {
				msg_b[j].dumpNoWrite();
				msg_f[i].dumpToDisk(SCRATCH_DIR+"/msgf"+itoc(i));
			}
			if (PAGE_MAPS && mer_i != mer) {
				prob_i->dumpNoWrite();
			}
			updatef[j-1]=false;
		}
		if(mer[3] != '_' && !fixRes[j] && !(fixRes[j+1] && count[j] > 1) && updateb[j+1]){
			int i = j+1;
			string mer_i = seqData.serialGet5mer(i);

			if (PAGE_MAPS) b[i].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
			b_i = b[i];
			if (PAGE_MAPS) b[i].dumpNoWrite();
			int ss_i = ss_to_index(sspred.serialGet(i));
			if(priorSoln[i]){
				mer_i=mer_i+itoc(i);
			}
			Belief* prob_i = &Prob0[ mer_i ];
			if (PAGE_MAPS && mer_i != mer) {
				prob_i->loadFromDisk(SCRATCH_DIR+"/p0_"+mer_i);
			}
			// page in
			if (PAGE_MAPS) {
					msg_f[j].loadFromDisk(SCRATCH_DIR+"/msgf"+itoc(j));
					msg_b[i].realloc();   // don't care what it is, just reallocate space
			}

			belief_subtract_and_proj(b_i, prob_i->phi_b, msg_f[j], ss_i);
			b_i.invlog_transform();
			compute_adj_message(msg_b[i], b_i, prob_i->phi_b, crystData, ss_i);
			msg_b[i].M.normalize(  M_EST_POSITIONS  );
			msg_b[i].M.log_transform();

			// page out
			if (PAGE_MAPS) {
				msg_f[j].dumpNoWrite();
				msg_b[i].dumpToDisk(SCRATCH_DIR+"/msgb"+itoc(i));
			}
			if (PAGE_MAPS && mer_i != mer) {
				prob_i->dumpNoWrite();
			}
			updateb[j+1]=false;
		}

		// aggregate incoming messages from adjacent residues
		if (mer[1] != '_' && !fixRes[j]) {
			if (PAGE_MAPS) {
				msg_glob[j-1].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j-1)+".dat");
				msg_f[j-1].loadFromDisk(SCRATCH_DIR+"/msgf"+itoc(j-1));
			}

			map_minus( b_j, b_j, msg_glob[j-1] );
			belief_add_and_proj ( b_j, prob_j.phi_b, msg_f[j-1], ss );

			// page out
			if (PAGE_MAPS) {
				msg_glob[j-1].dumpNoWrite();
				msg_f[j-1].dumpNoWrite();
			}
		}
		if (mer[3] != '_' && !fixRes[j]) {
			// page in
			if (PAGE_MAPS) {
				msg_glob[j+1].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j+1)+".dat");
				msg_b[j+1].loadFromDisk(SCRATCH_DIR+"/msgb"+itoc(j+1));
			}

			map_minus( b_j, b_j, msg_glob[j+1] );
			belief_add_and_proj ( b_j, prob_j.phi_f, msg_b[j+1] ,ss );

			// page out
			if (PAGE_MAPS) {
				msg_glob[j+1].dumpNoWrite();
				msg_b[j+1].dumpNoWrite();
			}
		}
		if (PAGE_MAPS) prob_j.dumpNoWrite();

		if(!fixRes[j]){// normalize belief
			b_j.logNormalize( M_EST_POSITIONS );
			//b_j.testIfContainsNaN("b_j");


			// compute change since last iteration
			if (PAGE_MAPS) b[j].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
			double this_delta = map_delta(b_j,b[j]);

			if (isnan(this_delta)) {
				cerr << "isNaN(delta[" << j << "]) == true!\n";
				cerr << "     " << b_j.sum() << " , " << b[j].sum() << " , "
					 << msg_glob[j-1].sum() <<  " , " << msg_glob[j].sum() <<  " , " << msg_glob[j+1].sum() << " , "
					 << msg_f[j-1].M.sum() << " , " << msg_b[j+1].M.sum() << endl;
				exit(1);
			}

			delta += this_delta;

			b[j] = b_j;
			if (PAGE_MAPS) b[j].dumpDataToDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
		}
		// occupancy message
		b_j.invlog_transform();
		if(!fixRes[j] || count[j]==2){ //Only compute global occupancy message once for fixed residues
			compute_occ_message(count[j], msg_glob[j], b_j, crystData);  // normalizes + log transforms msg
			//msg_glob[j].testIfContainsNaN("glob_j");

			map_plus ( msg_glob_prod, msg_glob_prod, msg_glob[j] );
		}
		if (PAGE_MAPS) msg_glob[j].dumpDataToDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");
		//if((iter/nResSum )% 1 == 0 && iter > 1){
//		for(int i = 0; i < nResSum; i++){
		if (doCheckPt && b[j].getNPts() > 0) {
			seqData.serialGet(j, resId, resType, chnId);
			if (PAGE_MAPS){
				b[j].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(j)+".dat");
				msg_glob[j].loadDataFromDisk(SCRATCH_DIR+"/msgglob"+itoc(j)+".dat");
				if(mer[3] != '_') msg_b[j+1].loadFromDisk(SCRATCH_DIR+"/msgb"+itoc(j+1));
				if(mer[1] != '_') msg_f[j-1].loadFromDisk(SCRATCH_DIR+"/msgf"+itoc(j-1));
			}
			b[j].writeDataToDisk(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal.dat"));
			msg_glob[j].writeDataToDisk(string(recoverdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_msg_glob.dat"));
			if(mer[3] != '_'){
				int resId2;
				char resType2, chnId2;
				seqData.serialGet(j+1, resId2, resType2, chnId2);
				msg_b[j+1].writeCheckPoint(string(recoverdir+filestem+"_chn"+chnId2+"_res"+itoc(resId2)+"_msg_b"));
			}
			if(mer[1] != '_'){
				int resId2;
				char resType2, chnId2;
				seqData.serialGet(j-1, resId2, resType2, chnId2);
				msg_f[j-1].writeCheckPoint(string(recoverdir+filestem+"_chn"+chnId2+"_res"+itoc(resId2)+"_msg_f"));
			}
			if (PAGE_MAPS){
				b[j].dumpNoWrite();
				msg_glob[j].dumpNoWrite();
				if(mer[3] != '_') msg_b[j+1].dumpNoWrite();
				if(mer[1] != '_') msg_f[j-1].dumpNoWrite();
			}
		//}
			msg_glob_prod.writeDataToDisk(string(recoverdir+filestem+"_msg_glob_prod.dat"));
			fstream miscfile((string(recoverdir+filestem+"_misc.txt")).c_str(),ios::out);
			miscfile << iter << endl;
			for(int i = 0; i < nResSum; i++){
				//updateb,updatef,count
				miscfile << updateb[i] << " " << updatef[i] << " " << count[i] << " "<< fixed << setprecision(7)  << linear_pq[i] << endl;
			}
			miscfile.close();
		}
	}
	cout << endl;
	cout << "Sum change in belief = " << delta << endl << endl;



	////////////////////
	//
	// calculate final trace, error
	//    - oversample the maps, and ensure adjacent CA's are 3.8A apart
	//    - also write final marginals to disk
	//
	doTrace (string(outpdbdir+filestem+"_final.pdb"), b, seqData, crystData,sspred);

	for (int i=0; i < nResSum; i++) {
		if (b[i].getNPts() > 0) {
			seqData.serialGet(i, resId, resType, chnId);

			if (PAGE_MAPS) b[i].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
			writeMapToDisk(b[i], string(outmapdir+filestem+"_chn"+chnId+"_res"+itoc(resId)+"_marginal"));
			if (PAGE_MAPS) b[i].dumpNoWrite();
		}
	}

	// clean up
	if (PAGE_MAPS) {
		DIR *pdir;
		struct dirent *pent;

		pdir=opendir(SCRATCH_DIR.c_str()); //"." refers to the current dir
		if (!pdir){
			printf ("opendir() failure; terminating");
			exit(1);
		}
		while ( pent = readdir(pdir) ){
			//cerr << "  deleting " << pent->d_name << endl;
			unlink( (SCRATCH_DIR+pent->d_name).c_str() );
		}
  		closedir(pdir);

  		rmdir(SCRATCH_DIR.c_str());
	}
	DIR *pdir;
	struct dirent *pent;
	pdir=opendir(recoverdir.c_str()); //"." refers to the current dir
	if (!pdir){
		printf ("opendir() failure; terminating");
		exit(1);
	}
	while ( pent = readdir(pdir) ){
		unlink( (recoverdir+pent->d_name).c_str() );
	}
	closedir(pdir);
	rmdir(recoverdir.c_str());
	//rename(recoverdir.c_str());
}
