// //////////////////////////////////////////////////////////////////////////
//
//     BP.h
//     (c) 2007 Frank DiMaio
//
// //////////////////////////////////////////////////////////////////////////
//
//     This file is part of ACMI.
//
//     ACMI is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 3 of the License, or
//     (at your option) any later version.
//
//     ACMI is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//     You should have received a copy of the GNU General Public License
//     along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// //////////////////////////////////////////////////////////////////////////


#include <string>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <map>

#include "prob_dist.h"
#include "crystal.h"
#include "compact_map3d.h"
#include "protein.h"

#include <ext/hash_map>
using __gnu_cxx::hash_map;

// model parameter
#ifndef INFINITY
#define INFINITY 1e30
#endif

// limit memory use
const double DEFAULT_MEM_LIMIT = 4*1024; // in MB (4 GB default)
const string DEFAULT_SCRATCH_DIR = "/scratch/"; // scratch (temp) directory

static bool  PAGE_MAPS;
const int NUM_SS = 4;
const double M_EST_PRIOR     = 0.1;     // spread this much prob. mass around uniformly INITIALLY
const double M_EST_POSITIONS = 0.001;   // spread this much prob. mass around uniformly AT EACH ITER.
const double R_COLLISION = 2.5;
const double CA_CA_DIST[] = {3.802525,3.802525,3.802525,3.802525};
const double CA_CA_SD[] = {0.05084,0.05,0.05,0.05};

// input params
static int    MAX_ITERS;
static double M_EST_ORIENT;
static double SIGMA;
static double MEM_LIMIT;
static string SCRATCH_DIR;

// cache
static vector< vec3<int> >        phi_adj_idxs[NUM_SS];
static vector< vec3<double> >     phi_adj_xs[NUM_SS];
static vector< double >           phi_adj_phis[NUM_SS];

static vector< vec3<int> >        phi_occ_idxs;

static hash_map< int , vector<float> > phi_cost_cache[NUM_SS];
static hash_map< int , vector<float> > phi_inv_cost_cache[NUM_SS];

// does 'filename' exist?
bool exists(string filename) {
	fstream fin;
	fin.open(filename.c_str(),ios::in);
	if( fin.is_open() )
		return true;
	return false;
}

// join the second vector to the end of the first
template <class T>
inline void append(vector<T> &X, vector<T> &Y) {
	int N=Y.size();
	for (int i=0; i<N; i++)
		X.push_back(Y[i]);
}

// building hashmaps keyed on string, vec3<int>
namespace __gnu_cxx {
	template<> class hash< vec3<int> >{
	public:
		size_t operator()(const vec3<int>& x) const { return x.hash(); }
	};
	template<> class hash<string>{
	public:
		size_t operator()(const string& x) const { return hash<const char*>()(x.c_str()); }
	};
}

int ss_to_index(char ss){
	switch(ss){
		case 'H' : return 1;
		case 'E' : return 2;
		case 'L' : return 3;
		case ' ' : return 0;
		default : return -1;
	}
}

/////////////////////////////////////
///
///  Build and Cache Potential Functions
///
///     * getAdjPotential
///     * getAdjPotentialList
///     * getFAdj
///     * getOccPotential
///     * getFOcc
///     * getFatAsymmUnit
///
/////////////////////////////////////

//Distance potential function
static map_3d< double >* getAdjPotential(vec3<int> dims , Crystal &crystData, int ss) {
	static map_3d< double >  map[NUM_SS];
	srand(1);

	if (map[ss].getDimsProd() == 0) { //Calculate potential once, cached afterwards
		if(map[ss].readFromFile("phiAdj"+itoc(ss))) return &map[ss];
		vec3<double> fX, fXovr;  // frac coords
		double pR;        // polar 'R'

		double K = max3(
			sqrt( vec3<double>(1.0/dims[0],0,0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,1.0/dims[1],0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,0,1.0/dims[2]).apply(crystData.f2c).norm2() )
		);//largest fractional grid size

		// use random sampling
		int NUM_SAMPLES = 100000;

		map[ss].alloc(dims);
		map[ss].setAll(0);

		for (int i=0; i<NUM_SAMPLES; ++i) { //sample possible adjacent residue locations
			vec3<double> fracX(
						((double)rand()/RAND_MAX-0.5)/dims[0],
						((double)rand()/RAND_MAX-0.5)/dims[1],
						((double)rand()/RAND_MAX-0.5)/dims[2]); // start with random coord

			double distXY = CA_CA_DIST[ss]+CA_CA_SD[ss]*randnorm();   // sample distance between residues

			// sample direction uniformly
			vec3<double> dirXY;
			double del;
			do {
				dirXY[0] = 2.0*(double)rand()/RAND_MAX-1.0;
				dirXY[1] = 2.0*(double)rand()/RAND_MAX-1.0;
				dirXY[2] = 2.0*(double)rand()/RAND_MAX-1.0;
				del = dirXY.norm2();
			} while (del >= 1.0 || del == 0);

			del = distXY / sqrt(del);
			dirXY = dirXY * del;

			vec3<double> fracY = fracX + dirXY.apply(crystData.c2f);
			vec3<int> idxY( pos_mod( (int)floor(fracY[0]*dims[0]+0.5) , dims[0]),
							pos_mod( (int)floor(fracY[1]*dims[1]+0.5) , dims[1]),
							pos_mod( (int)floor(fracY[2]*dims[2]+0.5) , dims[2]) );

			map[ss].inc(idxY[0],idxY[1],idxY[2], 1.0);
		}

		// ensure symmetry
		for (int i=0; i<dims[0]; i++)
		for (int j=0; j<dims[1]; j++)
		for (int k=0; k<dims[2]; k++) {
			fX[0] = ((double)i)/dims[0];  if (fX[0]>.5) fX[0] -=1;
			fX[1] = ((double)j)/dims[1];  if (fX[1]>.5) fX[1] -=1;
			fX[2] = ((double)k)/dims[2];  if (fX[2]>.5) fX[2] -=1;

			vec3<double> sX;
			vec3<int> sIdx;

			double map_ijk = map[ss].get(i,j,k);
			if (map_ijk == 0) continue;   // ???

			for (int s=1; s<crystData.nSymm(); ++s) {
				sX = fX.apply( crystData.getSymmRot(s) );
				sIdx[0] = pos_mod( (int)floor(sX[0]*dims[0]+0.5) , dims[0]);
				sIdx[1] = pos_mod( (int)floor(sX[1]*dims[1]+0.5) , dims[1]);
				sIdx[2] = pos_mod( (int)floor(sX[2]*dims[2]+0.5) , dims[2]);

				map_ijk = max(map_ijk, map[ss].get(sIdx[0],sIdx[1],sIdx[2]));
			}

			map[ss].set(i,j,k, map_ijk);
		}

		map[ss].normalize();  // sum to 1

		//writeMapToDisk(map[ss], "phiAdj"+itoc(ss));
	}

	return &map[ss];
}

//
// make adj potential list ... a list of all nonzero entries in the adj potential function
void getAdjPotentialList(vec3<int> dims, Crystal &crystData) {
	for(int ss=0; ss < NUM_SS; ss++){
		map_3d<double> *phi_adj = getAdjPotential(dims , crystData, ss);

		double phi_i;
		vec3<double> x_i;

		for (int i=0; i<dims[0]; i++) {
			x_i[0] = (double)i/dims[0];  if (x_i[0]>.5) x_i[0] -=1;
			for (int j=0; j<dims[1]; j++) {
				x_i[1] = (double)j/dims[1];  if (x_i[1]>.5) x_i[1] -=1;
				for (int k=0; k<dims[2]; k++) {
					x_i[2] = (double)k/dims[2];  if (x_i[2]>.5) x_i[2] -=1;

					phi_i = phi_adj->get(i,j,k);

					if (phi_i > 1e-6) {
						phi_adj_xs[ss].push_back( x_i.apply(crystData.f2c) );
						phi_adj_idxs[ss].push_back( vec3<int>(i,j,k) );
						phi_adj_phis[ss].push_back( phi_i );
					}
				}
			}
		}
	}
}

static map_3d< double >* getOccPotential(vec3<int> &dims , Crystal& crystData) {
	static map_3d< double > map;

	if (map.getDimsProd() == 0) {
		vec3<double> fX;  // frac coords
		double pR;        // polar 'R'
		double K = max3(
			sqrt( (vec3<double>(1.0/dims[0],0,0)).apply(crystData.f2c).norm2() ),
			sqrt( (vec3<double>(0,1.0/dims[1],0)).apply(crystData.f2c).norm2() ),
			sqrt( (vec3<double>(0,0,1.0/dims[2])).apply(crystData.f2c).norm2() )
		);
		double delR;

		map.alloc(dims);

		for (int i=0; i<dims[0]; i++) {
			fX[0] = (double)i/dims[0];  if (fX[0]>.5) fX[0] -=1;
			for (int j=0; j<dims[1]; j++) {
				fX[1] = (double)j/dims[1];  if (fX[1]>.5) fX[1] -=1;
				for (int k=0; k<dims[2]; k++) {
					fX[2] = (double)k/dims[2];  if (fX[2]>.5) fX[2] -=1;
					pR = sqrt(fX.apply(crystData.f2c).norm2());
					delR = pR - R_COLLISION;

					if (delR > K/2)
						map.set(i,j,k, 1.0);
					else if (delR < K/2 && delR > -K/2)
						map.set(i,j,k, (delR + K/2)/K );
					else
						map.set(i,j,k, 0.0);
				}
			}
		}
		//writeMapToDisk(map, "phiOcc");
	}

	return &map;
}

///
/// make adj potential list ... a list of all nonzero entries in the adj potential function
///
void getOccPotentialList(vec3<int> dims, Crystal &crystData) {
	vec3<double> fX;  // frac coords
	double pR;        // polar 'R'

	for (int i=0; i<dims[0]; i++) {
		fX[0] = (double)i/dims[0];  if (fX[0]>.5) fX[0] -=1;
		for (int j=0; j<dims[1]; j++) {
			fX[1] = (double)j/dims[1];  if (fX[1]>.5) fX[1] -=1;
			for (int k=0; k<dims[2]; k++) {
				fX[2] = (double)k/dims[2];  if (fX[2]>.5) fX[2] -=1;

				pR = sqrt(fX.apply(crystData.f2c).norm2());
				if( pR <= R_COLLISION ) {
					phi_occ_idxs.push_back( vec3<int>(i,j,k) );
				}

			}
		}
	}
}

///
///  return F[global potential] ... only compute it once
///
static map_3d< complex<double> >* getFOcc(vec3<int> &dims , Crystal& crystData) {
	static map_3d< complex<double> > Focc;
	if (Focc.getDimsProd() == 0) {
		fft( *getOccPotential( dims , crystData ) , Focc);
	}
	return &Focc;
}


///
/// make a "fixed potential" .. all prob mass at several (symmetric) locations
///      that is, we're certain of the location
///
void makeFixedPotential(map_3d<float> &map , vector< vec3<int> > xs, double M) {
	map.setAll(M);
	for (int i=0; i<xs.size(); ++i) {
		map.set(xs[i][0],xs[i][1],xs[i][2], 1.0);
	}
}


///
///  build a CA-Se potential from an PriorAtomlist
///
static compact_map_3d< double >* getSePotential(PriorAtomList &PAs, vec3<int> dims , Crystal &crystData) {
	static compact_map_3d< double > map;
	srand(1);

	if (map.getNPts() == 0) {
		int NUM_SAMPLES = 10000;

		vec3<double> fX, fXovr;  // frac coords
		double pR;        // polar 'R'

		double K = max3(
			sqrt( vec3<double>(1.0/dims[0],0,0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,1.0/dims[1],0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,0,1.0/dims[2]).apply(crystData.f2c).norm2() )
		);

		// use random sampling
		map.alloc(crystData, dims);
		map.setAll(0);

		int nPAs = PAs.nAtoms();

		for (int i=0; i<nPAs; ++i) {
			if(PAs[i].elt != 34) continue; //skip if not a Selenium
			for (int j=0; j<NUM_SAMPLES; ++j) {

				double distXY = rand()%10 < 6 ? 3.3+0.12*randnorm() : 4.15+0.04*randnorm();   // dist

				// direction
				vec3<double> dirXY;
				double del;
				do {
					dirXY[0] = 2.0*(double)rand()/RAND_MAX-1.0;
					dirXY[1] = 2.0*(double)rand()/RAND_MAX-1.0;
					dirXY[2] = 2.0*(double)rand()/RAND_MAX-1.0;
					del = dirXY.norm2();
				} while (del >= 1.0 || del == 0);

				del = distXY / sqrt(del);
				dirXY = dirXY * del;

				vec3<double> fracY = (PAs[i].cartX + dirXY).apply(crystData.c2f);
				vec3<int> idxY( pos_mod( (int)floor(fracY[0]*dims[0]+0.5) , dims[0]),
								pos_mod( (int)floor(fracY[1]*dims[1]+0.5) , dims[1]),
								pos_mod( (int)floor(fracY[2]*dims[2]+0.5) , dims[2]) );

				map.i_inc(idxY[0],idxY[1],idxY[2], 1.0);
			}
		}

		map.normalize();  // sum to 1
		//writeMapToDisk(map, "phiSe");
	}

	return &map;
}

///
///  build a CA-Se potential from an PriorAtomlist
///
static compact_map_3d< double >* getCAPotential(PriorAtomList &PAs, vec3<int> dims , Crystal &crystData) {
	static compact_map_3d< double > map;
	srand(1);

	if (map.getNPts() == 0) {
		int NUM_SAMPLES = 10000;

		vec3<double> fX, fXovr;  // frac coords
		double pR;        // polar 'R'

		double K = max3(
			sqrt( vec3<double>(1.0/dims[0],0,0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,1.0/dims[1],0).apply(crystData.f2c).norm2() ),
			sqrt( vec3<double>(0,0,1.0/dims[2]).apply(crystData.f2c).norm2() )
		);

		// use random sampling
		map.alloc(crystData, dims);
		map.setAll(0);

		int nPAs = PAs.nAtoms();

		for (int i=0; i<nPAs; ++i) {
			if(PAs[i].elt == 34) continue; //skip if not a Calpha
			for (int j=0; j<NUM_SAMPLES; ++j) {
				//TODO tune gaussian width, with sigma = 1, 25% of moving one grid point away
				double distXY = randnorm();   // dist

				// direction
				vec3<double> dirXY;
				double del;
				do {
					dirXY[0] = 2.0*(double)rand()/RAND_MAX-1.0;
					dirXY[1] = 2.0*(double)rand()/RAND_MAX-1.0;
					dirXY[2] = 2.0*(double)rand()/RAND_MAX-1.0;
					del = dirXY.norm2();
				} while (del >= 1.0 || del == 0);

				del = distXY / sqrt(del);
				dirXY = dirXY * del;

				vec3<double> fracY = (PAs[i].cartX + dirXY).apply(crystData.c2f);
				vec3<int> idxY( pos_mod( (int)floor(fracY[0]*dims[0]+0.5) , dims[0]),
								pos_mod( (int)floor(fracY[1]*dims[1]+0.5) , dims[1]),
								pos_mod( (int)floor(fracY[2]*dims[2]+0.5) , dims[2]) );

				map.i_inc(idxY[0],idxY[1],idxY[2], 1.0);
			}
		}

		map.normalize();  // sum to 1
		//writeMapToDisk(map, "phiCa");
	}

	return &map;
}

/////////////////////////////////////
///
///  struct Belief
///     a 3d grid of probabilities, together with a set of forward and backward orientations
///
///     a full 7-dimensional probability distribution is represented as the product of
///         the (interpolated) probability in space, WITH
///         the sum of gaussians centered at each forward (or backward) orientation
///
///     the number of "orientation" gaussians at each grid point is specified as a parameter
///     the stdev of the gaussians is globally SIGMA
///
class Belief {
public:
	Belief() { }


	static vec3<int> getDims(string fileFormatString, string mer) {
		map_3d<double> Mdouble;
		string filename = fileFormatString+"/match_"+mer+".cc.mat";
		if (Mdouble.readFromFile(filename.c_str()))
			return Mdouble.getDims();
		else
			return vec3<int>(0,0,0);
	}

	Belief(string fileFormatString, PriorAtomList &PAs, double SEprior, double CAprior, string mer, Crystal& crystData, int ss, bool priorS = false) {
		map_3d<double> beta,gamma,Mdouble;
		vec3<int> dims(0,0,0);

		////////////////
		// read CC maps (new version)
		//
		string filename = fileFormatString+"/match_"+mer+".cc.mat";
		int dimsProduct(0), index;
		if (Mdouble.readFromFile(filename.c_str())){
			// use original dimensions
			dims = Mdouble.getDims();

			dimsProduct = dims[0]*dims[1]*dims[2];
			beta.alloc(dims[0],dims[1],dims[2]);
			gamma.alloc(dims[0],dims[1],dims[2]);

			// convert CC -> Z-scores -> likelihoods
			if(!priorS){
				double meanCC = Mdouble.mean(),
			       stdCC  = sqrt(Mdouble.variance());
				double p_null;
				for (int i=0; i<dimsProduct; ++i) {
					p_null = normCDF( ( meanCC-Mdouble.get(i) ) / stdCC);
					Mdouble.set( i, (1.0-p_null)/p_null );
				}
			} else { //if prior structured provided, just add small probability around true location
				for (int i=0; i<dimsProduct; ++i) {
					Mdouble.set( i, Mdouble.get(i)+1e-20 );
				}
			}
		} else {
			cerr << "ERROR: " << filename << " cannot be read!\nCheck the working directory to ensure it contains 'priors' or rerun shfear";
		}

		// compact
		M.alloc(crystData, dims);
		for (int i=0; i<dimsProduct; ++i) {
			if (crystData.getAsymmUnit().get(i) > 0) {
				M.i_set(i, (float)Mdouble.get(i));
			}
		}
 		if(!priorS){
			M.normalize( M_EST_PRIOR );
 		} else {
 			M.normalize();
 		}

		// update based on Se locations?
		if (PAs.nSe() >0 && mer[2]=='M') {
			compact_map_3d<double> SE_prob;
			SE_prob.alloc( crystData , dims );

			for (int i=0; i<M.getNPts(); ++i) {
				SE_prob.d_set( i , getSePotential(PAs,dims,crystData)->d_get(i) * M.d_get(i) );
			}
			SE_prob.normalize( M_EST_PRIOR );

			// weighted sum
			for (int i=0; i<M.getNPts(); ++i) {
				M.d_set ( i , (float) ((1.0-SEprior) * M.d_get(i) + SEprior * SE_prob.d_get(i) ) );
			}
		}
		if (PAs.nCA() > 0){
			compact_map_3d<double> CA_prob;
			CA_prob.alloc( crystData , dims );
			for (int i=0; i<M.getNPts(); ++i) {
				CA_prob.d_set( i , getCAPotential(PAs,dims,crystData)->d_get(i) * M.d_get(i) );
			}
			CA_prob.normalize( M_EST_PRIOR );
			// weighted sum
			for (int i=0; i<M.getNPts(); ++i) {
				M.d_set ( i , (float) ((1.0-CAprior) * M.d_get(i) + CAprior * CA_prob.d_get(i) ) );
			}
		}
		// Forward orientations (if provided)
		if (mer[3] != '_') {
			// beta
			filename = fileFormatString+"/match_"+mer+".betaF.mat";
			if (!beta.readFromFile(filename.c_str())) {
				cerr << "WARNING: " << filename << " not found!\n";
				phi_f.clear(); phi_b.clear();
				return;
			}

			// gamma
			filename = fileFormatString+"/match_"+mer+".gammaF.mat";
			if (!gamma.readFromFile(filename.c_str())) {
				cerr << "WARNING: " << filename << " not found!\n";
				phi_f.clear(); phi_b.clear();
				return;
			}

			//
			phi_f.alloc(crystData, dims);
			//
			for (int j=0; j<dimsProduct; j++) {
				if (crystData.getAsymmUnit().get(j) == 0) continue;

				if (beta.get(j) < 0) {   // no orientation
					phi_f.i_set(j, 0);
				} else {
					vec3<double> x_j ( CA_CA_DIST[ss]*cos(gamma.get(j))*sin(beta.get(j)),
					                   CA_CA_DIST[ss]*sin(gamma.get(j))*sin(beta.get(j)),
					                   CA_CA_DIST[ss]*cos(beta.get(j)) );
					x_j = x_j.apply(crystData.c2f);  // convert to fractional

					// find closest grid point to beta[j],gamma[j] @ CA_CA_DIST angstrom
					vec3<int> idx_j( (int)floor(x_j[0]*dims[0]+0.5) ,
									 (int)floor(x_j[1]*dims[1]+0.5) ,
					                 (int)floor(x_j[2]*dims[2]+0.5) );
					idx_j.mod(dims);
					index = to_index(idx_j,dims);
					phi_f.i_set(j, index);

					if ( phi_cost_cache[ss].size() > 0 && phi_cost_cache[ss].count( index ) == 0 ) {
						cerr << "ERROR! " << filename << "\n";
						cerr << "     phi_cost_cache[" << gamma.get(j) << " , " << beta.get(j) << "] undefined\n";
						exit(1);
					}
				}
			}
		}

		// Backward orientations (if provided)
		if (mer[1] != '_') {
			// beta
			filename = fileFormatString+"/match_"+mer+".betaB.mat";
			if (!beta.readFromFile(filename.c_str())) {
				cerr << "WARNING: " << filename << " not found!\n";
				phi_f.clear(); phi_b.clear();
				return;
			}

			// gamma
			filename = fileFormatString+"/match_"+mer+".gammaB.mat";
			if (!gamma.readFromFile(filename.c_str())) {
				cerr << "WARNING: " << filename << " not found!\n";
				phi_f.clear(); phi_b.clear();
				return;
			}

			//
			phi_b.alloc(crystData, dims);
			//
			for (int j=0; j<dimsProduct; j++) {
				if (crystData.getAsymmUnit().get(j) == 0) continue;

				if (beta.get(j) < 0) {   // no orientation
					phi_b.i_set(j, 0);
				} else {
					vec3<double> x_j ( CA_CA_DIST[ss]*cos(gamma.get(j))*sin(beta.get(j)),
					                   CA_CA_DIST[ss]*sin(gamma.get(j))*sin(beta.get(j)),
					                   CA_CA_DIST[ss]*cos(beta.get(j)) );

					x_j = x_j.apply(crystData.c2f);  // convert to fractional

					vec3<int> idx_j( (int)floor(x_j[0]*dims[0]+0.5) ,
					                 (int)floor(x_j[1]*dims[1]+0.5) ,
					                 (int)floor(x_j[2]*dims[2]+0.5) );
					idx_j.mod(dims);
					index = to_index(idx_j,dims);
					phi_b.i_set(j, index);

					if (  phi_cost_cache[ss].size() > 0 && phi_cost_cache[ss].count( index ) == 0 ) {
						cerr << "ERROR! phi_cost_cache[" << index << " = phi_b.i_get(" << j << ")] undefined\n";
						exit(1);
					}
				}
			}
		}
	}

	void dumpToDisk(string filestem) {
		M.dumpDataToDisk(filestem + "_m.dat");
		if (phi_f.getNPts() > 0) phi_f.dumpDataToDisk(filestem + "_phif.dat");
		if (phi_b.getNPts() > 0) phi_b.dumpDataToDisk(filestem + "_phib.dat");
	}
	void dumpNoWrite() {
		M.dumpNoWrite();
		if (phi_f.getNPts() > 0) phi_f.dumpNoWrite();
		if (phi_b.getNPts() > 0) phi_b.dumpNoWrite();
	}
	void loadFromDisk(string filestem) {
		M.loadDataFromDisk(filestem + "_m.dat");
		if (phi_f.getNPts() > 0) phi_f.loadDataFromDisk(filestem + "_phif.dat");
		if (phi_b.getNPts() > 0) phi_b.loadDataFromDisk(filestem + "_phib.dat");
	}
	void realloc() {
		M.realloc();
		if (phi_f.getNPts() > 0) phi_f.realloc();
		if (phi_b.getNPts() > 0) phi_b.realloc();
	}

	compact_map_3d<float>  M;
	compact_map_3d<int>    phi_f, phi_b;
	bool fixedRes;
};

class AdjMessage {
public:
	void writeCheckPoint(string filestem){
		if(M.getNPts() > 0){
			M.writeDataToDisk(filestem+".M.dat");
			phi.writeDataToDisk(filestem+".phi.dat");
		}
	}
	void loadCheckPoint(string filestem){
		M.dumpNoWrite();
		phi.dumpNoWrite();
		M.loadDataFromDisk(filestem+".M.dat");
		phi.loadDataFromDisk(filestem+".phi.dat");
	}
	void dumpToDisk(string filestem) {
		M.dumpDataToDisk(filestem + "_m.dat");
		if (phi.getNPts() > 0) phi.dumpDataToDisk(filestem + "_phi.dat");
	}
	void dumpNoWrite() {
		M.dumpNoWrite();
		if (phi.getNPts() > 0) phi.dumpNoWrite();
	}
	void loadFromDisk(string filestem) {
		M.loadDataFromDisk(filestem + "_m.dat");
		if (phi.getNPts() > 0) phi.loadDataFromDisk(filestem + "_phi.dat");
	}
	void realloc() {
		M.realloc();
		if (phi.getNPts() > 0) phi.realloc();
	}

	compact_map_3d<float> M;
	compact_map_3d<int> phi;
};


/////////////////////////
///
///  Adjacency Messages
///
/// compute adj message (taking into account orientation information)
///      this requires for us to check all possible neighbor positions (specified in phi_adj_idxs)
///         and sum (alternately we could use max) probabilities (M_0, M_i)
///         WEIGHTED by proximity to our belief in the orientation (phis)
///      phi_cost_cache precomputes pairwise distances between phi_adj_idxs (in cartesian space)
///
///
/// use max-product
void compute_adj_message(AdjMessage               &msg,
						compact_map_3d<float>    &b,
						compact_map_3d<int>      &phi,
						Crystal                  &crystData, int ss) {
	vec3<int> dims = b.getDims();
	int xi,xj,xk, xidx, idx;
	int nneighbor = phi_adj_idxs[ss].size();
	float b_0, b_msg;

	map_3d<short> *asymmunit = &crystData.getAsymmUnit();
	map_3d<short> *asymmcopy = &crystData.getAsymmCopy();

	msg.M.alloc(crystData, dims);
	msg.M.setAll(0.0);

	compact_map_3d<float> Mmax;
	Mmax.alloc(crystData, dims);
	Mmax.setAll(0.0);

	msg.phi.alloc(crystData, dims);
	msg.phi.setAll(-1);

	xidx = -1;
	for (xi=0; xi<dims[0]; xi++)
	for (xj=0; xj<dims[1]; xj++)
	for (xk=0; xk<dims[2]; xk++) {
		xidx++;

		if (asymmunit->get(xidx) <= 0) continue; // if not in asymm unit no need to calculate (should save a lot of time!)
		b_0 = b.i_get(xidx);

		// POSSIBLE SPEEDUP here
		// if b_0 < some threshold, dont 'push' its belief to other locations

		if ( phi_cost_cache[ss].count( phi.i_get(xidx) ) == 0 ) {
			cerr << "ERROR! phi_cost_cache[" << phi.i_get(xidx) << " = phi.i_get(" << xidx << ") undefined\n";
			exit(1);
		}
		const vector<float> &cost_xi_0 = phi_cost_cache[ss][ phi.i_get(xidx) ];
		for (int k=0; k<nneighbor; k++) {
			const vec3<int> &phi_k = phi_adj_idxs[ss][k];

			vec3<int> srcPt( xi+phi_k[0] , xj+phi_k[1] , xk+phi_k[2]);
			srcPt.mod(dims);
			idx = to_index( srcPt , dims );

			//idx = ((xi+phi_k[0]+dims[0]) % dims[0]) * dims[1]*dims[2] +
			//      ((xj+phi_k[1]+dims[1]) % dims[1]) * dims[2] +
			//      ((xk+phi_k[2]+dims[2]) % dims[2]);

			b_msg =  b_0 * cost_xi_0[k] ;

			if (b_msg <= 0) {
				cout << " b<0! " << b_0 << " , " << cost_xi_0[k]  << " (" << k << "/" << nneighbor << ")\n";
				cout << "          " << xidx << " -- " << phi.i_get(xidx) << endl;
			}

			msg.M.i_inc(idx, b_msg);
			if (b_msg > Mmax.i_get(idx)) {
				Mmax.i_set(idx, b_msg);

				// IF msg crosses asymm unit boundary
				//    (that is, if idx is not in asymm unit)
				// then we need to adjust msg.phi
				// AS IF the message were coming from outside the asymmetric unit
				msg.phi.i_set( idx, (asymmcopy->get(idx)*nneighbor)+k );
			}
		}
	}
}


/////////////////////////////////////
//
// Adjacency message aggregation routines
//
/////////////////////////////////////
//
// belief_add_and_proj
//     given the projected belief b in directions phi,
//     add (mult. in real-space) incoming _directional_ message msg
//     return the projected belief
//
void belief_add_and_proj( compact_map_3d<float> &b, compact_map_3d<int> &phi, AdjMessage &msg, int ss ) {
	int nPts = b.getNPts();
	float toAdd;
	for (int i=0; i<nPts; ++i) {
		toAdd = msg.M.d_get(i);

		// test
		if ( phi_inv_cost_cache[ss].count( phi.d_get(i) ) == 0 ) {
			cerr << "ERROR! phi_inv_cost_cache[" << phi.d_get(i) << " = phi.d_get(" << i << ")] undefined\n";
			exit(1);
		}
		if (msg.phi.d_get(i) == -1) {
			cerr << "ERROR: msg.phi.d_get(" << i << ") == -1\n";
		}

		toAdd += phi_inv_cost_cache[ss][phi.d_get(i)][msg.phi.d_get(i)];
		b.d_inc(i , toAdd);
	}
}

//
// belief_subtract_and_proj
//     same as above, but subtract msg contribution
//
void belief_subtract_and_proj( compact_map_3d<float> &b, compact_map_3d<int> &phi, AdjMessage &msg, int ss ) {
	int nPts = b.getNPts();
	float toSub;
	for (int i=0; i<nPts; ++i) {
		toSub = msg.M.d_get(i);

		if ( phi_inv_cost_cache[ss].count( phi.d_get(i) ) == 0 ) {
			cerr << "ERROR! phi_inv_cost_cache[" << phi.d_get(i) << " = phi.d_get(" << i << ")] undefined\n";
			exit(1);
		}
		if (msg.phi.d_get(i) == -1) {
			cerr << "ERROR: msg.phi.d_get(" << i << ") == -1\n";
		}

		toSub += phi_inv_cost_cache[ss][phi.d_get(i)][msg.phi.d_get(i)];
		b.d_inc(i , -toSub);
	}
}


/////////////////////////////////////
//
// Occupancy Messages
//
// compute_occ_message
//    sum over the unit cell using a fourier convolution
//
void compute_occ_message(int iter,
                         compact_map_3d<float>            &msg,
                         compact_map_3d<float>             b,
                         Crystal                  &crystData) {
	if (iter == 1) {
		msg.setAll(0);
		return;
	}

	vec3<int> dims = b.getDims();
	int dimsProd = dims[0]*dims[1]*dims[2];

	// update res j's contribution to the global msg
	//      msg_glob[j] = ifftn( Fglobal.*fftn( exp(b{j}) ) );
	map_3d<double> b_full , msg_full;
	map_3d< complex<double> > fB, fMsg;
	b.inflate(b_full);

	crystData.maskAsymmetricUnit(b_full);
	fft(b_full,fB);
	map_times(fMsg, *getFOcc(dims,crystData), fB);
	ifft(fMsg, msg_full);
	msg_full.scale(1.0/dimsProd);   // fft, ifft are unscaled
	msg_full.log_transform();

	// prod cryst symm (sum of logs)
	crystData.sumSymmetric(msg_full);

	//
	msg_full.logNormalize( M_EST_POSITIONS );

	msg.alloc(crystData, dims);
	for (int i=0; i<dimsProd; i++) {
		if (crystData.getAsymmUnit().get(i) == 0) continue;
		msg.i_set(i, (float)msg_full.get(i));
	}

}

// oversample a map
inline void oversample(map_3d<double> &in, map_3d<double> &out, InterpolationType interp) {
	switch(interp) {
		case SPLINE2:
			resize_spline(in,out,2); break;
		case SPLINE3:
			resize_spline(in,out,3); break;
		case SPLINE4:
			resize_spline(in,out,4); break;
		case SPLINE5:
			resize_spline(in,out,5); break;
		case FFT:
			resize_fft(in,out); break;
		default:   // nearest
			resize_nn(in,out);
	}
}


//
//  find optimal trace given current belief
//    * oversample belief to 'dims' using 'interp' interpolation
//    * ensure local consistency
//    * output in PDB format to 'filename'
//
class traceNode {
public:
	vec3<double> fracX, cartX;
	vec3<int>    idxX;
	float        prob;
	int          resPtr;
};

class trace {
public:
	vec3<double>      cCoM, fCoM;
	vector<traceNode> res;
};

void doTrace(
	           string                            filename,
	           vector< compact_map_3d<float> >  &b,
	           Sequence                         &seqData,
	           Crystal                          &crystData,
	           Sequence							&ss
	          ) {
	int nRes = seqData.nres();

	// traces
	traceNode       thisRes, prevRes;
	trace           thisChain;
	vector< trace > protModel;

	///////////////////
	///
	/// A) extract max arg b_i for i=1...nres
	vector< vec3<double> > fsymm;
	for (int i=0; i < nRes; i++) {
		if (b[i].getNPts() != 0) {
			vec3<int> dims = b[i].getDims();

			if (PAGE_MAPS) b[i].loadDataFromDisk(SCRATCH_DIR+"/b"+itoc(i)+".dat");
			compact_map_3d<float> bCopy = b[i];
			if (PAGE_MAPS) b[i].dumpNoWrite();

			bCopy.invlog_transform();
			bCopy.argmax(thisRes.prob , thisRes.idxX); // compute ml label

			thisRes.fracX = vec3<double>((double)thisRes.idxX[0]/dims[0],
			                             (double)thisRes.idxX[1]/dims[1],
			                             (double)thisRes.idxX[2]/dims[2]);
			thisRes.resPtr = i;

			crystData.getSymmetric(thisRes.fracX, fsymm);

			if ( thisChain.res.size() != 0 ) {
				bool connectToPrev = false;
				for (int j=0; j<fsymm.size(); ++j) {
					vec3<int>  delIdxXY;
					vec3<double> delfXY;

					delfXY = fsymm[j] - thisChain.res.back().fracX;
					delfXY.minmod1();

					delIdxXY[0] = pos_mod( (int)(floor(thisChain.res.back().fracX[0]*dims[0]+0.5)
					                           - floor(fsymm[j][0]*dims[0]+0.5)) , dims[0]);
					delIdxXY[1] = pos_mod( (int)(floor(thisChain.res.back().fracX[1]*dims[1]+0.5)
					                           - floor(fsymm[j][1]*dims[1]+0.5)) , dims[1]);
					delIdxXY[2] = pos_mod( (int)(floor(thisChain.res.back().fracX[2]*dims[2]+0.5)
					                           - floor(fsymm[j][2]*dims[2]+0.5)) , dims[2]);


					// check if connected to previous
					if ( getAdjPotential(dims , crystData,ss_to_index(ss.serialGet(i)))->get(delIdxXY[0],delIdxXY[1],delIdxXY[2]) >= 1e-6 ) {
						// CONNECTED!
						connectToPrev = true;
						thisRes.fracX = thisChain.res.back().fracX + delfXY;
						thisRes.cartX = thisChain.res.back().cartX + delfXY.apply( crystData.f2c );
						//cerr << "Connect " << thisChain.res.back().cartX.to_string() << " to " << thisRes.cartX.to_string() << endl;
						break;

					}
				}

				if ( !connectToPrev ) {
					// NOT CONNECTED !
					if (thisChain.res.size() >= 3) {   // ENFORCE MIN CHAIN LENGTH!
						thisChain.cCoM = thisChain.cCoM / thisChain.res.size();   // compute CoM
						thisChain.fCoM = thisChain.cCoM.apply( crystData.c2f ); //thisChain.fCoM.mod(1);

						//cerr << "add chain (len=" << thisChain.res.size() << ") \n";
						//cerr << "   fcom = " << thisChain.fCoM.to_string() << " :: ";
						//cerr << "   ccom = " << thisChain.cCoM.to_string() << endl;

						//for (int k=0; k<thisChain.res.size(); ++k)
						//	cerr << "     "
						//	     << vec3<double>((double)thisChain.res[k].idxX[0]/dims[0],
						//		                 (double)thisChain.res[k].idxX[1]/dims[1],
						//						 (double)thisChain.res[k].idxX[2]/dims[2]).to_string()
						//	     << " :: "
						//	     << thisChain.res[k].fracX.to_string() << " :: " << thisChain.res[k].cartX.to_string() << endl;

						protModel.push_back(thisChain);
					}

					thisChain.res.clear();
					thisChain.cCoM = vec3<double>(0,0,0);

					thisRes.cartX = thisRes.fracX.apply( crystData.f2c );
				}
			} else {
				thisChain.cCoM = vec3<double>(0,0,0);
				thisRes.cartX = thisRes.fracX.apply( crystData.f2c );
			}

			thisChain.res.push_back( thisRes );
			thisChain.cCoM = thisChain.cCoM + thisRes.cartX;
		}
	}

	// add last chain
	if (thisChain.res.size() >= 3) {   // ENFORCE MIN CHAIN LENGTH!
		thisChain.cCoM = thisChain.cCoM / thisChain.res.size();   // compute CoM
		thisChain.fCoM = thisChain.cCoM.apply( crystData.c2f ); //thisChain.fCoM.mod(1);
		protModel.push_back(thisChain);

		//cerr << "add chain (len=" << thisChain.res.size() << ") \n";
		//cerr << "   fcom = " << thisChain.fCoM.to_string() << endl;
		//cerr << "   ccom = " << thisChain.cCoM.to_string() << endl;
	}


	if (protModel.size() == 0) { // no legal chains
		return;
	}

	///////////////////
	///
	/// B) move all to correct asymm unit
	int nchains = protModel.size();
	int movedCount=1;
	vector<bool> moved(nchains, false);
	int maxChainLen=protModel[0].res.size(), idxMaxChain=0;
	vec3<double> combinedCoM, sumCoM;

	for (int i=1; i<nchains; ++i) {
		if ( protModel[i].res.size() > maxChainLen ) {
			maxChainLen = protModel[i].res.size();
			idxMaxChain = i;
		}
	}
	moved[idxMaxChain] = true;
	combinedCoM = sumCoM = protModel[idxMaxChain].fCoM;

	//cerr << "chain " << idxMaxChain << " fixed\n";
	//cerr << "   com = " << protModel[idxMaxChain].fCoM.to_string() << endl;

	while (movedCount < nchains) {
		// for each chain, find distance closest to combinedCoM
		int idxMinDist, copyMinDist, idxResMinDist;
		double thisDist, minDist=INFINITY;
		vec3<double> minCoM;

		for (int i=0; i<nchains; ++i) {
			if (moved[i]) continue;

			crystData.getSymmetric(protModel[i].fCoM, fsymm);
			for (int j=0; j<fsymm.size(); ++j) {
				vec3<double> delfXY = fsymm[j] - combinedCoM;
				delfXY.minmod1();

				thisDist = delfXY.apply(crystData.f2c).norm2();

				if (thisDist < minDist) {
					minDist = thisDist;
					idxMinDist = i;
					copyMinDist = j;
					minCoM = combinedCoM + delfXY;
				}
			}
		}

		//cerr << "mindist = " << sqrt(minDist) << " fixed\n";
		//cerr << "   chain " << idxMinDist << endl;
		//cerr << "   symmcpy " << copyMinDist << endl;
		//cerr << "   new CoM = " << minCoM.to_string() << endl;


		// update CoM
		moved[idxMinDist] = true;
		movedCount++;
		sumCoM = sumCoM + minCoM;
		combinedCoM = sumCoM / movedCount;
 		protModel[idxMinDist].fCoM = minCoM;

		// move chain 'idxMinDist'
		int nres = protModel[idxMinDist].res.size();

		// find+set res closest to com
		minDist = INFINITY;
		for (int i=0; i<nres; ++i) {
			crystData.getSymmetric( protModel[idxMinDist].res[i].fracX,
			                        protModel[idxMinDist].res[i].fracX ,
									copyMinDist);

			vec3<double> delfXY = protModel[idxMinDist].res[i].fracX - minCoM;
			delfXY.minmod1();

			protModel[idxMinDist].res[i].fracX = minCoM + delfXY;

			thisDist = delfXY.apply(crystData.f2c).norm2();
			//cerr << i << ": thisRESdist = " << sqrt(thisDist) << "  ::  " << minCoM.to_string() << " :: "
			//                                                              << protModel[idxMinDist].res[i].fracX.to_string() << "\n";

			if (thisDist < minDist) {
				minDist = thisDist;
				idxResMinDist = i;
			}
		}
		//cerr << "minRESdist = " << sqrt(minDist) << "\n";
		//cerr << "   resID " << idxResMinDist << endl;
		//cerr << "   resPos = " << protModel[idxMinDist].res[idxResMinDist].fracX.to_string() << endl;
		//cerr << endl;

		protModel[idxMinDist].res[idxResMinDist].cartX =
		                  protModel[idxMinDist].res[idxResMinDist].fracX.apply( crystData.f2c );

		// move BW ...
		for (int i=idxResMinDist-1; i>=0; --i) {
			vec3<double> delfXY = protModel[idxMinDist].res[i].fracX - protModel[idxMinDist].res[i+1].fracX;
			delfXY.minmod1();

			//cerr << "interresdist  = " << sqrt( delfXY.apply( crystData.f2c ).norm2() ) << endl;

			protModel[idxMinDist].res[i].fracX = protModel[idxMinDist].res[i+1].fracX + delfXY;
			protModel[idxMinDist].res[i].cartX = protModel[idxMinDist].res[i].fracX.apply( crystData.f2c );
		}

		// and FW ...
		for (int i=idxResMinDist+1; i<nres; ++i) {
			vec3<double> delfXY = protModel[idxMinDist].res[i].fracX - protModel[idxMinDist].res[i-1].fracX;
			delfXY.minmod1();

			//cerr << "interresdist  = " << sqrt( delfXY.apply( crystData.f2c ).norm2() ) << endl;

			protModel[idxMinDist].res[i].fracX = protModel[idxMinDist].res[i-1].fracX + delfXY;
			protModel[idxMinDist].res[i].cartX = protModel[idxMinDist].res[i].fracX.apply( crystData.f2c );
		}
	}


	///////////////////
	///
	/// C) write to file
	ostream *pdbout;
	ofstream fileout;
	if (filename == "stdout")       pdbout = &cout;
	else if (filename == "stderr")  pdbout = &cerr;
	else {
		fileout.open(filename.c_str(), ios::out);
		pdbout = &fileout;
	}

	*pdbout << setiosflags(ios_base::right|ios::fixed) << setprecision(3);
	*pdbout << "REMARK  Generated automatically by BP.cpp\n";
	*pdbout << crystData.crystline << endl;;
	int thisAtmId = 0;
	for (int i=0; i<nchains; ++i) {
		int nres = protModel[i].res.size();
		for (int j=0; j<nres; ++j) {
			int resPtr = protModel[i].res[j].resPtr;
			*pdbout << "ATOM  " << setw(5) << ++thisAtmId << "  CA  "
			        << threeLetterCode(seqData.serialGet(resPtr)) << " "
					<< (char)('A'+i%26) << setw(4) << resPtr << "    ";
			*pdbout << setw(8) << protModel[i].res[j].cartX[0]
			        << setw(8) << protModel[i].res[j].cartX[1]
			        << setw(8) << protModel[i].res[j].cartX[2];
			*pdbout << "  1.00 " << protModel[i].res[j].prob << endl;

		}

		*pdbout << "TER   " << endl;
	}
}



